"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import TypeAlias

from .abc import Message, SteamID
from .chat import ChatMessage
from .reaction import Emoticon, MessageReaction, Sticker
from .utils import DateTime

if TYPE_CHECKING:
    from .channel import ClanChannel, DMChannel, GroupChannel
    from .clan import Clan
    from .group import Group
    from .protobufs import chat, friend_messages
    from .user import ClientUser, User


__all__ = (
    "UserMessage",
    "GroupMessage",
    "ClanMessage",
)

Authors: TypeAlias = "User | ClientUser | SteamID"


class UserMessage(Message):
    """Represents a message from a user."""

    channel: DMChannel
    mentions: None

    def __init__(self, proto: friend_messages.IncomingMessageNotification, channel: DMChannel):
        super().__init__(channel, proto)
        self.author = channel.participant
        self.created_at = DateTime.from_timestamp(proto.rtime32_server_timestamp)

    async def add_emoticon(self, emoticon: Emoticon) -> None:
        await self._state.react_to_user_message(
            self.author.id64,
            int(self.created_at.timestamp()),
            self.ordinal,
            emoticon.name,
            reaction_type=1,
            is_add=True,
        )
        self._state.dispatch(
            "reaction_add",
            MessageReaction(self._state, self, emoticon, None, self._state.user, DateTime.now(), self.ordinal),
        )

    async def remove_emoticon(self, emoticon: Emoticon):
        await self._state.react_to_user_message(
            self.author.id64,
            int(self.created_at.timestamp()),
            self.ordinal,
            emoticon.name,
            reaction_type=1,
            is_add=False,
        )
        self._state.dispatch(
            "reaction_remove",
            MessageReaction(self._state, self, emoticon, None, self._state.user, DateTime.now(), self.ordinal),
        )

    async def add_sticker(self, sticker: Sticker):
        await self._state.react_to_user_message(
            self.author.id64,
            int(self.created_at.timestamp()),
            self.ordinal,
            sticker.name,
            reaction_type=2,
            is_add=True,
        )
        self._state.dispatch(
            "reaction_add",
            MessageReaction(self._state, self, None, sticker, self._state.user, DateTime.now(), self.ordinal),
        )

    async def remove_sticker(self, sticker: Sticker):
        await self._state.react_to_user_message(
            self.author.id64,
            int(self.created_at.timestamp()),
            self.ordinal,
            sticker.name,
            reaction_type=2,
            is_add=False,
        )
        self._state.dispatch(
            "reaction_remove",
            MessageReaction(self._state, self, None, sticker, self._state.user, DateTime.now(), self.ordinal),
        )


class GroupMessage(ChatMessage):
    """Represents a message in a group."""

    channel: GroupChannel
    group: Group
    clan: None

    def __init__(self, proto: chat.IncomingChatMessageNotification, channel: GroupChannel, author: Authors):
        super().__init__(proto, channel, author)


class ClanMessage(ChatMessage):
    """Represents a message in a clan."""

    channel: ClanChannel
    clan: Clan
    group: None

    def __init__(self, proto: chat.IncomingChatMessageNotification, channel: ClanChannel, author: Authors):
        super().__init__(proto, channel, author)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from typing_extensions import TypeAlias

ID64: TypeAlias = int
ID32: TypeAlias = int

ChatGroupID: TypeAlias = int
ChannelID: TypeAlias = int



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import Any

from typing_extensions import Literal, TypedDict


class GetOwnedGames(TypedDict):
    games: list[GetOwnedGamesGame]


class GetOwnedGamesGame(TypedDict):
    name: str
    appid: str
    playtime_forever: int
    img_icon_url: str
    has_community_visible_stats: bool


class GameToDict(TypedDict, total=False):
    game_id: str
    game_extra_info: str


class WishlistGame(TypedDict):
    name: str
    capsule: str
    review_score: int
    review_desc: str
    reviews_total: str
    reviews_percent: int
    release_date: int
    release_string: str
    platform_icons: str
    subs: list[dict[str, Any]]
    type: str
    screenshots: list[str]
    review_css: str
    priority: int
    added: int
    background: str
    rank: int
    tags: list[str]
    is_free_game: bool
    win: bool
    mac: bool
    linux: bool


class PackageGroups(TypedDict):
    name: Literal["default", "subscriptions"]
    title: str
    description: str
    selection_text: str
    display_type: Literal[0, 1]
    is_recurring_subscription: bool
    subs: list[PackageGroupSub]


class PackageGroupSub(TypedDict):
    packageid: int
    percent_savings_text: str
    percent_savings: int
    option_text: str
    option_description: str
    can_get_free_license: str
    is_free_license: bool
    price_in_cents_with_discount: int


class FetchedGamePriceOverview(TypedDict):
    currency: str
    initial: int
    final: int
    discount_percent: int
    initial_formatted: str
    final_formatted: str


class FetchedGame(TypedDict):
    # https://wiki.teamfortress.com/wiki/User:RJackson/StorefrontAPI#Result_data_3
    type: Literal["game", "dlc", "demo", "advertising", "mod", "video"]
    name: str
    steam_appid: int
    required_age: int
    is_free: bool
    controller_support: Literal["partial", "full"]
    dlc: list[int]
    detailed_description: str
    short_description: str
    fullgame: dict[str, Any]
    supported_languages: str
    header_image: str
    pc_requirements: list[dict[str, str]]
    mac_requirements: list[dict[str, str]]
    linux_requirements: list[dict[str, str]]
    legal_notice: str
    developers: list[str]
    publishers: list[str]
    demos: list[dict[str, Any]]
    price_overview: FetchedGamePriceOverview
    packages: list[int]
    package_groups: list[PackageGroups]
    platforms: dict[str, bool]
    metacritic: list[dict[str, str]]
    categories: list[dict[str, str]]
    release_date: dict[str, str]
    background: str
    website: str
    movies: list[dict[str, Any]]


class DLCPriceOverview(TypedDict):
    currency: str
    initial: int
    final: int
    discount_percent: int


class DLCPlatforms(TypedDict):
    windows: bool
    mac: bool
    linux: bool


class DLCReleaseDate(TypedDict):
    steam: str  # unix timestamps
    mac: str
    linux: str


class DLC(TypedDict):
    id: int
    name: str
    header_image: str
    price_overview: DLCPriceOverview
    platforms: DLCPlatforms
    release_date: DLCReleaseDate



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import hashlib
import imghdr
import io
import struct
from time import time
from typing import TYPE_CHECKING, Any

from typing_extensions import Protocol

__all__ = ("Image",)

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath


class ImageIO(Protocol):
    def seekable(self) -> bool:
        ...

    def seek(self, __offset: int, __whence: int = ...) -> Any:
        ...

    def tell(self) -> int:
        ...

    def readable(self) -> bool:
        ...

    def read(self, __size: int = ...) -> bytes:
        ...

    def close(self) -> Any:
        ...


class Image:
    """A wrapper around common image files. Used for :meth:`steam.User.send`.

    Parameters
    ----------
    fp
        An image or path-like to pass to :func:`open`.
    spoiler
        Whether to mark the image as a spoiler.

    Note
    ----
    Currently supported image types include:
        - PNG
        - JPG/JPEG
        - GIF
    """

    # TODO add support for "webm", "mpg", "mp4", "mpeg", "ogv"

    __slots__ = ("fp", "spoiler", "name", "width", "height", "type", "hash", "size", "_tell")
    fp: ImageIO

    def __init__(self, fp: ImageIO | StrOrBytesPath | int, *, spoiler: bool = False):
        self.fp = fp if isinstance(fp, io.BufferedIOBase) else open(fp, "rb")
        if not (self.fp.seekable() and self.fp.readable()):
            raise ValueError(f"File buffer {fp!r} must be seekable and readable")

        self._tell = self.fp.tell()
        contents = self.read()
        self.size = len(contents)
        if self.size > 1024 * 1024 * 10:  # 10MiB
            raise ValueError("File is too large to upload")

        # from https://stackoverflow.com/questions/8032642
        headers = self.fp.read(24)
        if len(headers) != 24:
            raise ValueError("Opened file has no headers")
        self.type = imghdr.what(None, headers)
        if self.type == "png":
            check = struct.unpack(">i", headers[4:8])[0]
            if check != 0x0D0A1A0A:
                raise ValueError("Opened file's headers do not match a standard PNGs headers")
            width, height = struct.unpack(">ii", headers[16:24])
        elif self.type == "gif":
            width, height = struct.unpack("<HH", headers[6:10])
        elif self.type == "jpeg":
            try:
                self.fp.seek(self._tell)  # read 0xff next
                size = self._tell + 2
                ftype = 0
                while not 0xC0 <= ftype <= 0xCF or ftype in {0xC4, 0xC8, 0xCC}:
                    self.fp.seek(size, 1)
                    byte = self.fp.read(1)
                    while ord(byte) == 0xFF:
                        byte = self.fp.read(1)
                    ftype = ord(byte)
                    size = struct.unpack(">H", self.fp.read(2))[0] - 2
                # we are at a SOFn block
                self.fp.seek(1, 1)  # skip 'precision' byte.
                height, width = struct.unpack(">HH", self.fp.read(4))
            except Exception as exc:
                raise ValueError from exc
        else:
            raise TypeError("Unsupported file format passed")
        self.spoiler = spoiler
        self.width = width
        self.height = height
        self.hash = hashlib.sha1(contents).hexdigest()
        self.name = f'{int(time())}_{getattr(self.fp, "name", f"image.{self.type}")}'

    def read(self) -> bytes:
        self.fp.seek(self._tell)
        read = self.fp.read()
        self.fp.seek(self._tell)
        return read

    def __enter__(self) -> Image:
        return self

    def __exit__(self, *args: Any) -> None:
        self.fp.close()


def test_jpeg(h: bytes, _) -> str | None:  # adds support for more header types
    # SOI APP2 + ICC_PROFILE
    if h[:4] == "\xff\xd8\xff\xe2" and h[6:17] == b"ICC_PROFILE":
        return "jpeg"
    # SOI APP14 + Adobe
    if h[:4] == "\xff\xd8\xff\xee" and h[6:11] == b"Adobe":
        return "jpeg"
    # SOI DQT
    if h[:4] == "\xff\xd8\xff\xdb":
        return "jpeg"


imghdr.tests.append(test_jpeg)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import re
import warnings
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, NamedTuple, TypeVar, overload

from typing_extensions import Literal

from . import utils
from ._const import DOCS_BUILDING, URL
from .enums import AppFlag, Enum, Language, PublishedFileQueryFileType, PublishedFileRevision, ReviewType
from .iterators import GamePublishedFilesIterator, GameReviewsIterator, ManifestIterator
from .utils import DateTime, Intable, id64_from_url

if TYPE_CHECKING:
    from .clan import Clan
    from .friend import Friend
    from .manifest import Depot, GameInfo, HeadlessDepot, Manifest
    from .package import FetchedGamePackage
    from .protobufs import player
    from .review import Review
    from .state import ConnectionState
    from .types import game
    from .user import User

__all__ = (
    "Game",
    "TF2",
    "LFD2",
    "DOTA2",
    "CSGO",
    "STEAM",
    "CUSTOM_GAME",
    "DLC",
    "UserGame",
    "WishlistGame",
    "FetchedGame",
)

T = TypeVar("T")
APP_ID_MAX = 2**32 - 1


class Game:
    """Represents a Steam game.

    Attributes
    ----------
    name
        The game's name.
    id
        The game's app ID.
    context_id
        The context id of the game normally ``2``.
    """

    __slots__ = (
        "id",
        "name",
        "context_id",
    )

    @overload
    def __init__(self, *, id: Intable, name: str | None = ..., context_id: int | None = ...):
        ...

    @overload
    def __init__(self, *, name: Literal["Team Fortress 2"], id: Intable = ..., context_id: int | None = ...):
        ...

    @overload
    def __init__(self, *, name: Literal["Left 4 Dead 2"], id: Intable = ..., context_id: int | None = ...):
        ...

    @overload
    def __init__(self, *, name: Literal["DOTA 2"], id: Intable = ..., context_id: int | None = ...):
        ...

    @overload
    def __init__(
        self, *, name: Literal["Counter Strike Global-Offensive"], id: Intable = ..., context_id: int | None = ...
    ):
        ...

    @overload
    def __init__(self, *, name: Literal["Steam"], id: Intable = ..., context_id: int | None = ...):
        ...

    def __init__(
        self,
        *,
        id: Intable | None = None,
        name: str | None = None,
        title: str | None = None,
        context_id: int | None = None,
    ):
        if title is not None:
            warnings.warn("Game.title is depreciated, use Game.name instead", DeprecationWarning)
        name = name or title
        if name is None and id is None:
            raise TypeError("__init__() missing a required keyword argument: 'id' or 'name'")

        if id is None:
            game = utils.get(Games, name=name)
            if game is None:
                raise ValueError(f"Cannot find a matching game for {name!r}")
            id = game.id
        else:
            try:
                id = int(id)
            except (ValueError, TypeError):
                raise ValueError("id expected to support int()") from None

            game = utils.get(Games, id=id)
            if game is not None:
                name = game.name

        if id < 0:
            raise ValueError("id cannot be negative")

        if name == "Steam" and context_id is None:
            context_id = 6

        self.id: int = id
        self.name: str | None = name
        self.context_id: int = 2 if context_id is None else context_id

    def __str__(self) -> str:
        return self.name or ""

    def __repr__(self) -> str:
        attrs = ("name", "id", "context_id")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"{self.__class__.__name__}({', '.join(resolved)})"

    def __eq__(self, other: Any) -> bool:
        return self.id == other.id if isinstance(other, Game) else NotImplemented

    def __hash__(self) -> int:
        return hash(self.id)

    @property
    def title(self) -> str | None:
        """The game's name.

        .. deprecated:: 0.8.0

            Use :attr:`name` instead.
        """
        warnings.warn("Game.title is depreciated, use Game.name instead", DeprecationWarning, stacklevel=2)
        return self.name

    def to_dict(self) -> game.GameToDict:
        if self.is_steam_game():
            return {"game_id": str(self.id)}

        if self.name is None or self.id is None:
            raise TypeError("un-serializable game with no title and or id")
        return {"game_id": str(self.id), "game_extra_info": self.name}

    def is_steam_game(self) -> bool:
        """Whether the game could be a Steam game."""
        return self.id <= APP_ID_MAX

    @property
    def url(self) -> str:
        """What should be the game's url on steamcommunity if applicable."""
        return f"{URL.COMMUNITY}/app/{self.id}"


class Games(Game, Enum):
    """This is "enum" to trick type checkers into allowing Literal[TF2] to be valid for overloads in extensions."""

    __slots__ = ("_name",)

    def __new__(cls, name: str, *args: Any, value: tuple[str, int, int] | tuple[()] = ()) -> Games:
        self = object.__new__(cls)
        set_attribute = object.__setattr__

        if args:  # being called when docs are building
            name = ""
            value = (name, args[0], args[1])

        assert value
        set_attribute(self, "_name", name)
        set_attribute(self, "name", value[0])
        set_attribute(self, "id", value[1])
        set_attribute(self, "context_id", value[2])
        return self

    if DOCS_BUILDING:

        def __init__(self, *args: Any) -> None:
            ...

    def __repr__(self) -> str:
        return self._name

    @property
    def value(self) -> int:
        return self.id

    TF2 = "Team Fortress 2", 440, 2
    LFD2 = "Left 4 Dead 2", 550, 2
    DOTA2 = "DOTA 2", 570, 2
    CSGO = "Counter Strike Global-Offensive", 730, 2
    STEAM = "Steam", 753, 6


TF2 = Games.TF2
DOTA2 = Games.DOTA2
CSGO = Games.CSGO
LFD2 = Games.LFD2
STEAM = Games.STEAM


@overload
def CUSTOM_GAME(name: str) -> Game:  # type: ignore
    ...


def CUSTOM_GAME(name: str | None = None, title: str | None = None) -> Game:
    """Create a custom game instance for :meth:`~steam.Client.change_presence`.
    The :attr:`Game.id` will be set to ``15190414816125648896`` and the :attr:`Game.context_id` to ``None``.

    Example:

    .. code-block:: python3

        await client.change_presence(game=steam.CUSTOM_GAME("my cool game"))

    Parameters
    ----------
    name
        The name of the game to set your playing status to.
    """
    if name is None and title is None:
        raise TypeError("CUSTOM_GAME missing argument 'name'")
    return Game(name=name or title, id=15190414816125648896, context_id=None)


class FriendThoughts(NamedTuple):
    recommended: list[Friend]
    not_recommended: list[Friend]


class StatefulGame(Game):
    """Games that have state."""

    __slots__ = ("_state",)

    def __init__(self, state: ConnectionState, **kwargs: Any):
        super().__init__(**kwargs)
        self._state = state

    def __repr__(self) -> str:
        attrs = ("name", "id", "context_id")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    async def clan(self) -> Clan:
        """Fetch this game's clan.

        This can be useful to get a Game's updates.

        .. code-block:: python3

            clan = await game.clan()
            async for update in clan.announcements().filter(
                lambda announcement: announcement.type in (steam.ClanEvent.MajorUpdate, steam.ClanEvent.SmallUpdate)
            ):
                ...  # do something with the update

        Raises
        ------
        ValueError
            This game has no associated clan.
        """

        id64 = await id64_from_url(self.url, self._state.http._session)
        if id64 is None:
            raise ValueError("Game has no associated clan")
        clan = await self._state.fetch_clan(id64)
        if clan is None:
            raise ValueError("Game has no associated clan")
        return clan

    async def player_count(self) -> int:
        """The games current player count."""
        return await self._state.fetch_game_player_count(self.id)

    async def friends_who_own(self) -> list[Friend]:
        """Fetch the users in your friend list who own this game."""
        id64s = await self._state.fetch_friends_who_own(self.id)
        return [self._state.get_friend(id64) for id64 in id64s]

    async def review(
        self,
        content: str,
        *,
        recommend: bool,
        public: bool = True,
        commentable: bool = True,
        received_compensation: bool = False,
        language: Language | None = None,
    ) -> Review:
        """Review a game.

        Parameters
        ----------
        content
            The content of the review.
        recommend
            Whether you recommended the game.
        public
            Whether the review should be public.
        commentable
            Whether the review should allow comments.
        received_compensation
            Whether you received compensation for this review.
        language
            The language the review is in.
        """
        language = language or self._state.http.language
        await self._state.http.post_review(
            self.id, content, recommend, public, commentable, received_compensation, language.api_name
        )
        return await self._state.user.fetch_review(self)  # TODO this sucks can we actually get the id ourselves?

    def reviews(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> GameReviewsIterator:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a :class:`steam.Game`'s
        :class:`steam.Review`\\s.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for review in game.reviews(limit=10):
                print("Reviewer:", review.author)
                print("Said:", review.content)

        Flattening into a list:

        .. code-block:: python3

            reviews = await game.reviews(limit=50).flatten()
            # reviews is now a list of Review

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of reviews to search through. Default is ``100``. Setting this to ``None`` will fetch all
            the game's reviews, but this will be a very slow operation.
        before
            A time to search for reviews before.
        after
            A time to search for reviews after.

        Yields
        ------
        :class:`~steam.Review`
        """
        return GameReviewsIterator(self._state, self, limit, before, after)

    async def friend_thoughts(self) -> FriendThoughts:
        """Fetch the client user's friends who recommended and didn't recommend this game in a review.

        .. source:: FriendThoughts
        """
        proto = await self._state.fetch_friend_thoughts(self.id)
        return FriendThoughts(
            [self._state.get_friend(utils.make_id64(id)) for id in proto.accountids_recommended],
            [self._state.get_friend(utils.make_id64(id)) for id in proto.accountids_not_recommended],
        )

    # async def fetch(self) -> Self & FetchedGame:  # TODO update signature to this when types.Intersection is done
    #     fetched = await self._state.client.fetch_game(self)
    #     return utils.update_class(fetched, copy.copy(self))

    async def fetch(self, *, language: Language | None = None) -> FetchedGame:
        """Shorthand for:

        .. code-block:: python3

            game = await client.fetch_game(game)
        """
        game = await self._state.client.fetch_game(self, language=language)
        if game is None:
            raise ValueError("Fetched game was not valid.")
        return game

    async def depots(self) -> Sequence[Depot | HeadlessDepot]:
        info = await self.info()
        return await info.depots()

    async def fetch_manifest(
        self, *, id: int, depot_id: int, branch: str = "public", password_hash: str = ""
    ) -> Manifest:
        """Fetch a CDN manifest for a game.

        Parameters
        ----------
        id
            The ID of the manifest to fetch.
        depot_id
            The ID of the manifest's associated depot.
        """
        return await self._state.fetch_manifest(
            self.id, id, depot_id, name=None, branch=branch, password_hash=password_hash
        )

    def manifests(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
        branch: str = "public",
        password: str | None = None,
        password_hash: str = "",
    ) -> ManifestIterator:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a :class:`steam.Game`'s
        :class:`steam.Manifest`\\s.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for manifest in game.manifests(limit=10):
                print("Manifest:", manifest.name)
                print(f"Contains {len(manifest.paths)} manifests")

        Flattening into a list:

        .. code-block:: python3

            manifests = await game.manifests(limit=50).flatten()
            # manifests is now a list of Manifest

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of :class:`.Manifests` to return.
        before
            The time to get manifests before.
        after
            The time to get manifests after.
        branch
            The name of the branch to fetch manifests from.
        password
            The password for the branch, if any.
        password_hash
            The hashed password for a manifest.

        Yields
        ------
        :class:`Manifest`
        """
        return ManifestIterator(
            state=self._state,
            limit=limit,
            before=before,
            after=after,
            game=self,
            branch=branch,
            password=password,
            password_hash=password_hash,
        )

    def published_files(
        self,
        *,
        type: PublishedFileQueryFileType = PublishedFileQueryFileType.Items,
        revision: PublishedFileRevision = PublishedFileRevision.Default,
        language: Language | None = None,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> GamePublishedFilesIterator:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a game's :class:`steam.PublishedFile`\\s.

        Examples
        --------

        Usage:

        .. code-block:: python3

            async for published_file in game.published_files(limit=10):
                print("Published file:", published_file.name)
                print("Published at:", published_file.created_at)
                print("Published by:", published_file.author)

        Flattening into a list:

        .. code-block:: python3

            published_files = await game.published_files(limit=50).flatten()
            # published_files is now a list of PublishedFile

        All parameters are optional.

        Parameters
        ----------
        type
            The type of published files to fetch.
        revision
            The desired revision of the published files to fetch.
        language
            The language to fetch the published files in. If ``None`` the current language is used.
        limit
            The maximum number of published files to search through. Default is ``100``. Setting this to ``None`` will
            fetch all the game's published files, but this will be a very slow operation.
        before
            A time to search for published files before.
        after
            A time to search for published files after.

        Yields
        ------
        :class:`~steam.PublishedFile`
        """
        return GamePublishedFilesIterator(self._state, self, type, revision, language, limit, before, after)

    async def info(self) -> GameInfo:
        """Shorthand for:

        .. code-block:: python3

            (info,) = await client.fetch_product_info(games=[game])
        """
        (info,), _ = await self._state.fetch_product_info((self.id,))
        return info

    async def dlc(self, *, language: Language | None = None) -> list[DLC]:
        """Fetch the game's DLC.

        Parameters
        ----------
        language
            The language to fetch the DLC in. If ``None``, the current language will be used.
        """
        data = await self._state.http.get_game_dlc(self.id, language)
        self.name = data["name"]
        return [DLC(self._state, dlc) for dlc in data["dlc"]]

    async def packages(self, *, language: Language | None = None) -> list[FetchedGamePackage]:
        """Fetch the game's packages.

        Parameters
        ----------
        language
            The language to fetch the packages in. If ``None``, the current language will be used.
        """
        fetched = await self.fetch(language=language)
        return fetched._packages

    async def add_free_licenses(self) -> list[License]:
        """Request the free licenses for this game.

        Raises
        ------
        ValueError
            No licenses were granted.
        """
        _, licenses = await self._state.request_free_license(self.id)
        return licenses


@dataclass
class PartialGamePriceOverview:
    currency: str
    initial: int
    final: int
    discount_percent: int


class DLC(StatefulGame):
    """Represents DLC (downloadable content) for a game.

    Attributes
    ----------
    created_at
        The time the DLC was released at.
    logo_url
        The logo url of the DLC.
    price_overview
        A price overview for the DLC.
    """

    name: str

    __slots__ = (
        "created_at",
        "logo_url",
        "price_overview",
        "_on_windows",
        "_on_mac_os",
        "_on_linux",
    )

    def __init__(self, state: ConnectionState, data: game.DLC):
        super().__init__(state, id=data["id"], name=data["name"])
        self.created_at = DateTime.from_timestamp(int(data["release_date"]["steam"]))
        self.logo_url: str = data["header_image"]
        self.price_overview = PartialGamePriceOverview(**data["price_overview"])

        platforms = data["platforms"]
        self._on_windows: bool = platforms["windows"]
        self._on_mac_os: bool = platforms["mac"]
        self._on_linux: bool = platforms["linux"]

    def is_free(self) -> bool:
        """Whether the game is free to download."""
        return not self.price_overview.final

    def is_on_windows(self) -> bool:
        """Whether the game is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the game is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the game is playable on Linux."""
        return self._on_linux


class UserGame(StatefulGame):
    """Represents a Steam game fetched by :meth:`steam.User.games`

    Attributes
    ----------
    playtime_forever
        The total time the game has been played for.
    icon_url
        The icon url of the game.
    playtime_two_weeks
        The amount of time the user has played the game in the last two weeks.
    playtime_windows
        The total amount of time the user has played the game on Windows.
    playtime_mac_os
        The total amount of time the user has played the game on macOS.
    playtime_linux
        The total amount of time the user has played the game on Linux.
    last_played_at
        The time the user last played this game at.
    """

    __slots__ = (
        "playtime_forever",
        "playtime_two_weeks",
        "playtime_windows",
        "playtime_mac_os",
        "playtime_linux",
        "icon_url",
        "last_played_at",
        "_stats_visible",
    )

    name: str

    def __init__(self, state: ConnectionState, proto: player.GetOwnedGamesResponseGame):
        super().__init__(state=state, id=proto.appid, name=proto.name)
        self.playtime_forever: timedelta = timedelta(minutes=proto.playtime_forever)
        self.playtime_two_weeks: timedelta = timedelta(minutes=proto.playtime_2_weeks)
        self.playtime_windows: timedelta = timedelta(minutes=proto.playtime_windows_forever)
        self.playtime_mac_os: timedelta = timedelta(minutes=proto.playtime_mac_forever)
        self.playtime_linux: timedelta = timedelta(minutes=proto.playtime_linux_forever)
        self.icon_url = (
            f"https://cdn.cloudflare.steamstatic.com/steamcommunity/public/images/apps/{self.id}/"
            f"{proto.img_icon_url}.jpg"
        )
        self.last_played_at = DateTime.from_timestamp(proto.rtime_last_played)

        self._stats_visible = proto.has_community_visible_stats

    def has_visible_stats(self) -> bool:
        """Whether the game has publicly visible stats."""
        return self._stats_visible


class WishlistGame(StatefulGame):
    """Represents a Steam game fetched by :meth:`steam.User.wishlist`\\.

    Attributes
    ----------
    priority
        The priority of the game in the wishlist.
    added_at
        The time that the game was added to their wishlist.
    created_at
        The time the game was uploaded at.
    background_url
        The background URL of the game.
    rank
        The global rank of the game by popularity.
    review_status
        The review status of the game.
    score
        The score of the game out of ten.
    screenshots
        The screenshots of the game.
    tags
        The tags of the game.
    total_reviews
        The total number reviews for the game.
    type
        The type of the app.
    logo_url
        The logo url of the game.
    """

    __slots__ = (
        "priority",
        "added_at",
        "background_url",
        "created_at",
        "logo_url",
        "rank",
        "review_status",
        "score",
        "screenshots",
        "tags",
        "total_reviews",
        "type",
        "_free",
        "_on_linux",
        "_on_mac_os",
        "_on_windows",
    )

    name: str

    def __init__(self, state: ConnectionState, id: int | str, data: game.WishlistGame):
        super().__init__(state, id=id, name=data["name"])
        self.logo_url = data["capsule"]
        self.score = data["review_score"]
        self.total_reviews = int(data["reviews_total"].replace(",", ""))
        self.review_status = ReviewType[data["review_desc"].replace(" ", "")]
        self.created_at = DateTime.from_timestamp(int(data["release_date"]))
        self.type: str = data["type"]
        self.screenshots = [
            f"https://cdn.cloudflare.steamstatic.com/steam/apps/{self.id}/{screenshot_url}"
            for screenshot_url in data["screenshots"]
        ]
        self.added_at = DateTime.from_timestamp(data["added"])
        self.background_url = data["background"]
        self.tags = data["tags"]
        self.rank = data["rank"]
        self.priority = int(data["priority"])

        self._free = data["is_free_game"]
        self._on_windows = bool(data.get("win", False))
        self._on_mac_os = bool(data.get("mac", False))
        self._on_linux = bool(data.get("linux", False))

    def is_free(self) -> bool:
        """Whether the game is free to download."""
        return self._free

    def is_on_windows(self) -> bool:
        """Whether the game is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the game is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the game is playable on Linux."""
        return self._on_linux


class Movie:
    __slots__ = ("name", "id", "url", "created_at")

    def __init__(self, movie: dict[str, Any]):
        self.name: str = movie["name"]
        self.id: int = movie["id"]
        self.url: str = movie["mp4"]["max"]
        match = re.search(r"t=(\d+)", self.url)
        self.created_at = DateTime.from_timestamp(int(match[1])) if match else None

    def __repr__(self) -> str:
        attrs = ("name", "id", "url", "created_at")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<Movie {' '.join(resolved)}>"


@dataclass
class GamePriceOverview(PartialGamePriceOverview):
    initial_formatted: str
    final_formatted: str


class FetchedGame(StatefulGame):
    """Represents a Steam game fetched by :meth:`steam.Client.fetch_game`\\.

    Attributes
    ----------
    created_at
        The time the game was uploaded at.
    background_url
        The background URL of the game.
    type
        The type of the app.
    logo_url
        The logo URL of the game.
    partial_dlc
        The game's downloadable content.
    website_url
        The website URL of the game.
    developers
        The developers of the game.
    publishers
        The publishers of the game.
    description
        The short description of the game.
    full_description
        The full description of the game.
    movies
        A list of the game's movies, each of which has ``name``\\, ``id``\\, ``url`` and optional
        ``created_at`` attributes.
    price_overview
        The price overview of the game.
    """

    __slots__ = (
        "logo_url",
        "background_url",
        "created_at",
        "type",
        "partial_dlc",
        "_packages",
        "website_url",
        "developers",
        "publishers",
        "description",
        "full_description",
        "movies",
        "price_overview",
        "_free",
        "_on_windows",
        "_on_mac_os",
        "_on_linux",
    )

    name: str

    def __init__(self, state: ConnectionState, data: game.FetchedGame):
        super().__init__(state, id=data["steam_appid"], name=data["name"])
        self.logo_url = data["header_image"]
        self.background_url = data["background"]
        self.created_at = (
            DateTime.parse_steam_date(data["release_date"]["date"], full_month=False)
            if data["release_date"]["date"]
            else None
        )
        self.type = AppFlag.from_str(data["type"])
        self.price_overview = GamePriceOverview(**data["price_overview"])

        self.partial_dlc = [StatefulGame(state, id=dlc_id) for dlc_id in data.get("dlc", [])]

        from .package import FetchedGamePackage

        self._packages = [
            FetchedGamePackage(state, package)
            for package_group in data["package_groups"]
            for package in package_group["subs"]
        ]

        self.website_url = data.get("website")
        self.developers = data["developers"]
        self.publishers = data["publishers"]
        self.description = data["short_description"]
        self.full_description = data["detailed_description"]

        self.movies = [Movie(movie) for movie in data["movies"]] if "movies" in data else None

        self._free = data["is_free"]
        self._on_windows = bool(data["platforms"].get("windows", False))
        self._on_mac_os = bool(data["platforms"].get("mac", False))
        self._on_linux = bool(data["platforms"].get("linux", False))

    async def packages(self, *, languages: Language | None = None) -> list[FetchedGamePackage]:
        return self._packages

    def is_free(self) -> bool:
        """Whether the game is free to download."""
        return self._free

    def is_on_windows(self) -> bool:
        """Whether the game is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the game is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the game is playable on Linux."""
        return self._on_linux



"""Licensed under The MIT License (MIT) - Copyright (c) 2020 James H-B. See LICENSE"""

from typing_extensions import NotRequired, Required, TypedDict


class AssetToDict(TypedDict):
    assetid: str
    amount: int
    appid: str
    contextid: str


class Asset(AssetToDict):
    instanceid: str
    classid: str
    missing: bool
    # rollback_new_assetid: NotRequired[str]


class ItemDescriptionLine(TypedDict):
    type: str
    value: str
    color: str
    label: str


class ItemAction(TypedDict):
    link: str
    name: str


class ItemTag(TypedDict):
    appid: int
    category: str
    internal_name: str
    localized_category_name: str
    localized_tag_name: str
    color: str


class Description(TypedDict, total=False):
    instanceid: Required[str]
    classid: Required[str]
    market_name: str
    currency: int
    name: str
    market_hash_name: str
    name_color: str
    background_color: str  # hex code
    type: str
    descriptions: list[ItemDescriptionLine]
    owner_descriptions: list[ItemDescriptionLine]
    actions: list[ItemAction]
    owner_actions: list[ItemAction]
    market_actions: list[ItemAction]
    tags: list[ItemTag]
    icon_url: str
    icon_url_large: str
    tradable: bool  # 1 vs 0
    marketable: bool  # same as above
    commodity: int  # might be a bool
    fraudwarnings: list[str]


class Item(Asset, Description):
    """We combine Assets with their matching Description to form items."""


class Inventory(TypedDict):
    assets: list[Asset]
    descriptions: list[Description]
    total_inventory_count: int
    success: int  # Result
    rwgrsn: int  # p. much always -2


class TradeOffer(TypedDict):
    tradeofferid: str
    tradeid: str  # only used for receipts (it's not the useful one)
    accountid_other: int
    message: str
    trade_offer_state: int  # TradeOfferState
    expiration_time: int  # unix timestamps
    time_created: int
    time_updated: int
    escrow_end_date: int
    items_to_give: list[Item]
    items_to_receive: list[Item]
    is_our_offer: bool
    from_real_time_trade: bool
    confirmation_method: int  # https://cs.github.com/SteamDatabase/SteamTracking/blob/e86f560898e9f8fbc93fa4f55d5872b03db5f72b/Structs/enums.steamd#L1607


class GetTradeOffer(TypedDict):
    offer: TradeOffer
    descriptions: list[Description]


class TradeOfferReceiptAsset(Asset):
    new_assetid: str
    new_contextid: str


class TradeOfferReceiptItem(TradeOfferReceiptAsset, Item):
    pass


class TradeOfferReceipt(TypedDict):
    status: int
    tradeid: str
    time_init: int
    assets_received: NotRequired[list[TradeOfferReceiptAsset]]
    assets_given: NotRequired[list[TradeOfferReceiptAsset]]
    descriptions: list[Description]


class TradeOfferHistoryTrade(TypedDict):
    tradeid: str
    steamid_other: str
    # message: str
    time_init: int
    status: int
    assets_given: list[TradeOfferReceiptAsset]
    assets_received: list[TradeOfferReceiptAsset]


class GetTradeOfferHistory(TypedDict):
    more: bool
    trades: list[TradeOfferHistoryTrade]
    descriptions: list[Description]


class TradeOfferCreateResponse(TypedDict):
    tradeofferid: str
    needs_mobile_confirmation: bool
    needs_email_confirmation: bool
    email_domain: str



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING

from . import utils
from .badge import Badge
from .enums import (
    Language,
    ProfileCustomisationStyle,
    ProfileItemClass,
    ProfileItemType,
    PublishedFileRevision,
    Result,
    UserBadge,
)
from .errors import WSException
from .game import StatefulGame
from .trade import Asset, Item

if TYPE_CHECKING:
    from .abc import BaseUser
    from .protobufs import player
    from .published_file import PublishedFile
    from .state import ConnectionState
    from .types import trade

__all__ = (
    "ProfileInfo",
    "ProfileItem",
    "OwnedProfileItems",
    "EquippedProfileItems",
    "ProfileShowcaseSlot",
    "ProfileCustomisation",
    "ProfileShowcase",
    "Profile",
)


@dataclass
class ProfileInfo:
    """Represents the user's profile info."""

    created_at: datetime
    """The time at which the user was created at."""
    real_name: str | None
    """The real name of the user."""
    city_name: str | None
    """The city the user is located in."""
    state_name: str | None
    """The name of the state the user is located in."""
    country_name: str | None
    """The name of the country the user is located in."""
    headline: str | None
    """The profile's headline."""
    summary: str
    """The user's summary."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} real_name={self.real_name!r}>"


@dataclass
class ProfileMovie:
    url: str  # TODO add more attributes like maybe created_at?


class ProfileItem:
    """Represents an item on/in a user's profile."""

    __slots__ = (
        "id",
        "url",
        "name",
        "title",
        "description",
        "game",
        "type",
        "class_",
        "movie",
        "equipped_flags",
        "owner",
        "_state",
        "_um_name",
    )

    def __init__(
        self, state: ConnectionState, owner: BaseUser, item: player.ProfileItem, *, um_name: str | None = None
    ):
        self.id = item.communityitemid
        """The item's id."""
        self.url = item.image_large
        """The item's url."""
        self.name = item.name
        """The item's name."""
        self.title = item.item_title
        """The item's title."""
        self.description = item.item_description
        """The item's description."""
        self.game = StatefulGame(state, id=item.appid)
        """The game the item is from."""
        self.type = ProfileItemType.try_value(item.item_type)
        """The item's type."""
        self.class_ = ProfileItemClass.try_value(item.item_class)
        """The item's class."""
        self.movie = ProfileMovie(item.movie_mp4)
        """The movie associated with the item."""
        self.equipped_flags = item.equipped_flags  # TODO might be useful for item show case?
        """The item's equipped flags."""

        self.owner = owner
        self._state = state
        self._um_name = um_name

    def __repr__(self) -> str:
        return f"<ProfileItem id={self.id} name={self.name!r} game={self.game!r}>"

    async def equip(self) -> None:
        """Equip the profile item."""
        if self._um_name is None:
            raise ValueError(f"Cannot equip {self!r}")

        msg = await self._state.ws.send_um_and_wait(f"Player.Set{self._um_name}", communityitemid=self.id)
        if msg.result != Result.OK:
            raise WSException(msg)

    async def item(self, *, language: Language | None = None) -> Item:
        """Resolve this to an actual item in the owner's inventory."""
        inventory = await self.owner.inventory(self.game, language=language)
        item = utils.get(inventory, id=self.id)
        assert item
        return item


@dataclass
class OwnedProfileItems:
    r"""Represents the :class:`ClientUser`\'s owned items."""

    backgrounds: list[ProfileItem]
    """The backgrounds the client user owns."""
    mini_profile_backgrounds: list[ProfileItem]
    """The mini profile backgrounds the client user owns."""
    avatar_frames: list[ProfileItem]
    """The avatar frames the client user owns."""
    animated_avatars: list[ProfileItem]
    """The animated avatars the client user owns."""
    modifiers: list[ProfileItem]
    """The modifiers the client user owns."""


@dataclass
class EquippedProfileItems:
    """Represents the items the user has equipped."""

    background: ProfileItem | None
    """The equipped background."""
    mini_profile_background: ProfileItem | None
    """The equipped mini profile background for the user."""
    avatar_frame: ProfileItem | None
    """The equipped avatar frame for the user."""
    animated_avatar: ProfileItem | None
    """The equipped animated avatar for the user."""
    modifier: ProfileItem | None
    """The equipped modifier for the user."""


@dataclass
class ProfileShowcaseSlot:
    __slots__ = (
        "_state",
        "owner",
        "name",
        "content",
        "index",
        "game",
        "asset",
        "published_file_id",
        "badge_id",
        "border_colour",
    )
    _state: ConnectionState
    owner: BaseUser
    name: str | None
    """The slot's name."""
    content: str | None
    """ The slot's description."""
    index: int | None
    """The slot's index."""
    game: StatefulGame
    """The slot's associated game."""

    asset: Asset | None
    """The :class:`Asset` the slot is associated with."""
    published_file_id: int | None
    """The ID of the :class:`PublishedFile` the slot is associated with."""
    badge_id: UserBadge | None
    """The ID of the :class:`UserBadge` the slot is associated with."""
    border_colour: int | None
    """The border colour of the slot."""

    async def item(self, *, language: Language | None = None) -> Item:
        """Fetches the associated :class:`.Item` from :attr:`asset`.

        Parameters
        ----------
        language
            The language to fetch the item in. If ``None``, the current language is used.
        """
        if self.asset is None:
            raise ValueError
        key = (self.asset.class_id, self.asset.instance_id)
        resp = await self._state.http.get_item_info(self.game.id, [key], language)
        data: trade.Item = {
            **resp[key],
            **self.asset.to_dict(),
            "missing": False,
        }  # type: ignore  # I don't wanna type out this in full to make this type-safe
        return Item(self._state, data, self.owner)

    async def published_file(
        self, *, revision: PublishedFileRevision = PublishedFileRevision.Default, language: Language | None = None
    ) -> PublishedFile:
        """Fetches the associated :class:`.PublishedFile` from :attr:`published_file_id`.

        Parameters
        ----------
        revision
            The revision of the published file to fetch.
        language
            The language to fetch the published file in. If ``None``, the current language is used.
        """
        if self.published_file_id is None:
            raise ValueError
        (file,) = await self._state.fetch_published_files((self.published_file_id,), revision, language)
        assert file
        return file

    async def badge(self) -> Badge:
        """Fetches the associated :class:`.Badge` from :attr:`badge_id`."""
        if self.badge_id is None:
            raise ValueError
        badges = await self.owner.badges()
        badge = utils.get(badges.badges, id=self.badge_id)
        assert badge
        return badge


@dataclass
class ProfileShowcase:
    __slots__ = (
        "_state",
        "owner",
        "type",
        "large",
        "active",
        "purchase_id",
        "level",
        "style",
        "slots",
    )
    _state: ConnectionState
    owner: BaseUser
    type: ProfileItemType
    """The showcase type."""
    large: bool
    """Whether the showcase is large."""
    active: bool
    """Whether the showcase is active."""
    purchase_id: int
    """The purchase id."""
    level: int
    """The level of the the showcase."""
    style: ProfileCustomisationStyle
    """The style of the showcase."""
    slots: list[ProfileShowcaseSlot]
    """The slots in this showcase."""

    async def items(self, *, language: Language | None) -> list[Item]:
        """Fetches the associated :class:`.Item`s for the entire showcase.

        Parameters
        ----------
        language
            The language to fetch the items in. If ``None``, the current language is used.
        """
        asset_map: defaultdict[int, dict[tuple[int, int], Asset]] = defaultdict(dict)

        for slot in self.slots:
            if slot.asset:
                asset_map[slot.asset._app_id][(slot.asset.class_id, slot.asset.instance_id)] = slot.asset

        return [
            Item(self._state, {**description, **assets[key].to_dict(), "missing": False}, self.owner)
            for app_id, assets in asset_map.items()
            for key, description in (await self._state.http.get_item_info(app_id, assets, language)).items()
        ]

    async def published_file(
        self, *, revision: PublishedFileRevision = PublishedFileRevision.Default, language: Language | None = None
    ) -> PublishedFile:
        """Fetches the associated :class:`.PublishedFile` from :attr:`published_file_id`.

        Parameters
        ----------
        revision
            The revision of the file to fetch.
        language
            The language to fetch the file in. If ``None``, the current language is used.
        """
        return await self.slots[0].published_file(revision=revision)

    async def badges(self) -> list[Badge]:
        """Fetches the associated :class:`.Badge`s for the entire showcase."""
        all_badges = await self.owner.badges()
        badges = [utils.get(all_badges.badges, id=slot.badge_id) for slot in self.slots if slot.badge_id]
        assert all(badges)
        return badges  # type: ignore


@dataclass
class ProfileTheme:
    __slots__ = ("id", "name")
    id: str
    name: str


@dataclass
class PurchasedCustomisation:  # TODO should this be merged into ProfileCustomisationSlot.purchase_id if the ids match
    __slots__ = ("id", "type", "level")
    id: int
    type: ProfileCustomisationStyle
    level: int


class ProfileCustomisation:
    """Represents a user's profile customisations."""

    __slots__ = (
        "showcases",
        "slots_available",
        "profile_theme",
        "purchased_customisations",
        "owner",
        "_state",
    )

    def __init__(self, state: ConnectionState, user: BaseUser, proto: player.GetProfileCustomizationResponse):
        self.showcases = [
            ProfileShowcase(
                state,
                owner=user,
                slots=[
                    ProfileShowcaseSlot(
                        state,
                        owner=user,
                        index=slot.slot or None,
                        game=StatefulGame(state, id=slot.appid),
                        published_file_id=slot.publishedfileid or None,
                        name=slot.title or None,
                        content=slot.notes or None,
                        badge_id=UserBadge.try_value(slot.badgeid) if slot.badgeid else None,
                        border_colour=slot.border_color or None,
                        asset=Asset(
                            {
                                "assetid": slot.item_assetid,  # type: ignore  # this just gets cast to an int anyway
                                "appid": slot.appid,
                                "contextid": slot.item_contextid,
                                "instanceid": slot.item_instanceid,
                                "classid": slot.item_classid,
                                "amount": 1,
                                "missing": False,
                            },
                            user,
                        )
                        if slot.item_assetid
                        else None,
                    )
                    for slot in customisation.slots
                ],
                type=ProfileItemType.try_value(customisation.customization_type),
                large=customisation.large,
                active=customisation.active,
                style=ProfileCustomisationStyle.try_value(customisation.customization_style),
                purchase_id=customisation.purchaseid,
                level=customisation.level,
            )
            for customisation in proto.customizations
        ]
        self.slots_available = proto.slots_available
        self.profile_theme = ProfileTheme(proto.profile_theme.theme_id, proto.profile_theme.title)
        self.purchased_customisations = [
            PurchasedCustomisation(
                customisation.purchaseid,
                ProfileCustomisationStyle.try_value(customisation.customization_type),
                customisation.level,
            )
            for customisation in proto.purchased_customizations
        ]
        self.owner = user
        self._state = state

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} owner={self.owner!r}>"


class Profile(EquippedProfileItems, ProfileCustomisation):
    """Represents a user's complete profile."""

    __slots__ = ()

    def __init__(self, equipped_items: EquippedProfileItems, customisation_info: ProfileCustomisation):
        utils.update_class(equipped_items, self)
        utils.update_class(customisation_info, self)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} owner={self.owner!r}>"


class FriendProfile(ProfileInfo, EquippedProfileItems, ProfileCustomisation):
    """Represents a friend's complete profile."""

    def __init__(
        self, equipped_items: EquippedProfileItems, info: ProfileInfo, customisation_info: ProfileCustomisation
    ):
        utils.update_class(equipped_items, self)
        utils.update_class(info, self)
        utils.update_class(customisation_info, self)


class ClientUserProfile(FriendProfile):
    """Represents a :class:`ClientUser`'s full profile."""

    __slots__ = ("owned_items",)

    def __init__(
        self,
        equipped_items: EquippedProfileItems,
        info: ProfileInfo,
        customisation_info: ProfileCustomisation,
        items: OwnedProfileItems,
    ):
        super().__init__(equipped_items, info, customisation_info)
        self.owned_backgrounds = items.backgrounds
        """The backgrounds the client user owns."""
        self.owned_mini_profile_backgrounds = items.mini_profile_backgrounds
        """The mini profile backgrounds the client user owns."""
        self.owned_avatar_frames = items.avatar_frames
        """The avatar frames the client user owns."""
        self.owned_animated_avatars = items.animated_avatars
        """The animated avatars the client user owns."""
        self.owned_modifiers = items.modifiers
        """The modifiers the client user owns."""



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from collections.abc import Coroutine
from typing import TYPE_CHECKING, Any

from ...abc import Channel, Message, Messageable

if TYPE_CHECKING:
    from datetime import datetime

    from ...clan import Clan
    from ...group import Group
    from ...image import Image
    from ...iterators import AsyncIterator
    from ...state import ConnectionState
    from ...user import User
    from .bot import Bot
    from .cog import Cog
    from .commands import Command
    from .utils import Shlex


__all__ = ("Context",)


class Context(Messageable["Message"]):
    """Represents the context of a command.

    Attributes
    ----------
    message
        The message the context was generated from.
    prefix
        The prefix of the message the context was generated from
    author
        The author of the message.
    channel
        The channel the message was sent in.
    clan
        The clan the message was sent in ``None`` if the message wasn't sent in a clan.
    group
        The group the message was sent in ``None`` if the message wasn't sent in a group.
    bot
        The bot instance.
    command
        The command the context is attached to.
    cog
        The cog the command is in.
    """

    def __init__(self, **attrs: Any):
        self.bot: Bot = attrs["bot"]
        self.message: Message = attrs["message"]
        self.lex: Shlex = attrs["lex"]
        self.prefix: str | None = attrs["prefix"]

        self.command: Command | None = attrs.get("command")
        self.cog: Cog | None = self.command.cog if self.command is not None else None
        self.invoked_with: str | None = attrs.get("invoked_with")

        self.author: User = self.message.author
        self.channel: Channel[Message] = self.message.channel
        self.clan: Clan | None = self.message.clan
        self.group: Group | None = self.message.group
        self._state: ConnectionState = self.message._state

        self.args: tuple[Any, ...] | None = None
        self.kwargs: dict[str, Any] | None = None
        self.command_failed: bool = False

    def _message_func(self, content: str) -> Coroutine[Any, Any, Message]:
        return self.channel._message_func(content)

    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        return self.channel._image_func(image)

    async def invoke(self) -> None:
        """A shortcut method that invokes the current context using :meth:`~steam.ext.commands.Command.invoke`.

        Equivalent to

        .. code-block:: python3

            await ctx.command.invoke(ctx)
        """
        await self.command.invoke(self)

    def history(
        self,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncIterator[Message]:
        """A shortcut method that gets the current channel's history.

        Equivalent to

        .. code-block:: python3

            ctx.channel.history(**kwargs)
        """
        return self.channel.history(limit=limit, before=before, after=after)

    @property
    def valid(self) -> bool:
        """Whether or not the context could be invoked."""
        return self.prefix is not None and self.command is not None



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import time
from typing import TYPE_CHECKING, Generic, TypeVar, Union

from ...enums import IntEnum
from .errors import CommandOnCooldown

if TYPE_CHECKING:
    from ...abc import Message, Messageable

__all__ = (
    "BucketType",
    "Cooldown",
)
T_Bucket = TypeVar("T_Bucket", "tuple[int, ...]", int)


class BucketType(IntEnum):
    """A enumeration that handles cooldowns.

    Each attribute operates on a per x basis. So :attr:`User` operates on a per :class:`~steam.User` basis.
    """

    # fmt: off
    Default = 0  #: The default :class:`BucketType` this operates on a global basis.
    User    = 1  #: The :class:`BucketType` for a :class:`steam.User`.
    Member  = 2  #: The :class:`BucketType` for a :class:`steam.User` and a :class:`steam.Clan`.
    Group   = 3  #: The :class:`BucketType` for a :class:`steam.User` in a :class:`steam.Clan` / :class:`steam.Group`.
    Clan    = 4  #: The :class:`BucketType` for a :class:`steam.Clan`.
    Role    = 5  #: The :class:`BucketType` for a :class:`steam.Role`.
    Channel = 6  #: The :class:`BucketType` for a :class:`steam.Channel`.
    Admin   = 7  #: The :class:`BucketType` for a :class:`steam.Clan`'s :attr:`steam.Clan.admins`.
    # fmt: on

    def get_bucket(self, ctx: Message | Messageable) -> int | tuple[int, ...]:
        """Get a bucket key for a message or context.

        Parameters
        ----------
        ctx
            The message or context to get the bucket for.
        """
        if self == BucketType.Default:
            return 0
        elif self == BucketType.User:
            return ctx.author.id
        elif self == BucketType.Member:
            return (ctx.clan and ctx.clan.id), ctx.author.id
        elif self == BucketType.Group:
            return (ctx.group or ctx.author).id
        elif self == BucketType.Role:
            return (ctx.clan and ctx.author.top_role.id), ctx.author.id
        elif self == BucketType.Clan:
            return (ctx.clan or ctx.author).id
        elif self == BucketType.Channel:
            return (ctx.channel or ctx.author).id
        elif self == BucketType.Admin:
            return (ctx.clan and (ctx.author in ctx.clan.admins)), ctx.author.id


class Cooldown(Generic[T_Bucket]):
    """The class that holds a command's cooldown."""

    bucket: BucketType  #: The bucket that should be used to determine this command's cooldown.

    def __init__(self, rate: int, per: float, bucket: BucketType):
        self._rate = rate
        self._per = per
        self.bucket = bucket
        self.reset()

    def reset(self) -> None:
        """Reset the command's cooldown."""
        self._queue: dict[float, T_Bucket] = {}

    def _calls(self, bucket: T_Bucket) -> list[tuple[float, T_Bucket], ...]:
        return [(t, b) for t, b in self._queue.items() if b == bucket]

    async def expire_cache(self, bucket: T_Bucket, now: float) -> None:
        self._queue[now] = bucket
        await asyncio.sleep(self._per)
        self._queue.pop(now, None)

    def get_retry_after(self, bucket: T_Bucket, now: float) -> float:
        """Get the retry after for a command.

        Parameters
        ----------
        bucket
            The bucket to find in the cache.
        now
            The UNIX timestamp to find times after.
        """
        calls = self._calls(bucket)
        if not calls:
            return 0.0
        last_call = calls[0][0]
        if last_call + self._per >= now and len(calls) >= self._rate:
            return last_call + self._per - now
        return 0.0

    def __call__(self, ctx: Message | Messageable) -> None:
        """Invoke the command's cooldown properly and raise if the command is on cooldown.

        Parameters
        ----------
        ctx
            The context for invocation to check for a cooldown on.
        """
        bucket = self.bucket.get_bucket(ctx)
        now = time.time()
        asyncio.create_task(self.expire_cache(bucket, now))
        retry_after = self.get_retry_after(bucket, now)
        if retry_after:
            raise CommandOnCooldown(retry_after)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from typing_extensions import TypedDict


class User(TypedDict):
    steamid: str
    personaname: str
    primaryclanid: str
    profileurl: str
    realname: str
    # enums
    communityvisibilitystate: int
    profilestate: int
    commentpermission: int
    personastate: int
    personastateflags: int
    # avatar
    avatar: str
    avatarmedium: str
    avatarfull: str
    avatarhash: str
    # country stuff
    loccountrycode: str
    locstatecode: int
    loccityid: int
    # game stuff
    gameid: str  # game app id
    gameextrainfo: str  # game name
    # unix timestamps
    timecreated: int
    lastlogoff: int
    # we pass these ourselves
    last_logon: int
    last_seen_online: int
    rich_presence: dict[str, str]



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import warnings
from collections.abc import Callable
from datetime import timedelta
from ipaddress import IPv4Address
from typing import TYPE_CHECKING, Any, Generic, NamedTuple, TypeVar, overload

from typing_extensions import Literal, TypeAlias, Unpack

from .abc import SteamID
from .enums import Enum, GameServerRegion, Type
from .game import Game, StatefulGame
from .protobufs.game_servers import EQueryType, GetServerListResponseServer, QueryResponse

if TYPE_CHECKING:
    from .state import ConnectionState

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
Q: TypeAlias = "Query[Q]"

__all__ = (
    "Query",
    "GameServer",
    "ServerPlayer",
)


class Operator(Enum):
    # fmt: off
    div  = "\\"
    nor  = "\\nor\\"
    nand = "\\nand\\"
    # fmt: on

    def format(self, query_1: str, query_2: str) -> str:
        return f"{query_1}{query_2}" if self is Operator.div else f"{self.value}[{query_1}{query_2}]"


class QueryAll:
    def __repr__(self) -> str:
        return self.__class__.__name__

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__)

    query = ""


class QueryMeta(type):
    @property
    def not_empty(cls) -> Q:
        """Fetches servers that are not empty."""
        return Query["Q"](r"\empty\1")

    @property
    def empty(cls) -> Q:
        """Fetches servers that are empty."""
        return Query["Q"](r"\noplayers\1")

    @property
    def proxy(cls) -> Q:
        """Fetches servers that are spectator proxies."""
        return Query["Q"](r"\proxy\1")

    @property
    def whitelisted(cls) -> Q:
        """Fetches servers that are whitelisted."""
        return Query["Q"](r"\white\1")

    @property
    def dedicated(cls) -> Q:
        """Fetches servers that are running dedicated."""
        return Query["Q"](r"\dedicated\1")

    @property
    def secure(cls) -> Q:
        """Fetches servers that are using anti-cheat technology (VAC, but potentially others as well)."""
        return Query["Q"](r"\secure\1")

    @property
    def linux(cls) -> Q:
        """Fetches servers running on a Linux platform."""
        return Query["Q"](r"\linux\1")

    @property
    def no_password(cls) -> Q:
        """Fetches servers that are not password protected."""
        return Query["Q"](r"\password\0")

    @property
    def not_full(cls) -> Q:
        """Fetches servers that are not full."""
        return Query["Q"](r"\full\1")

    @property
    def unique_addresses(cls) -> Q:
        """Fetches only one server for each unique IP address matched."""
        return Query["Q"]("\\collapse_addr_hash\\1")

    @property
    def version_match(cls) -> Query[str]:
        """Fetches servers running version "x" (``"*"`` is wildcard)."""
        return Query["str"]("\\version_match\\", type=str)

    @property
    def name_match(cls) -> Query[str]:
        """Fetches servers with their hostname matching "x" (``"*"`` is wildcard)."""
        return Query["str"]("\\name_match\\", type=str)

    @property
    def running_mod(cls) -> Query[str]:
        """Fetches servers running the specified modification (e.g. cstrike)."""
        return Query["str"]("\\gamedir\\", type=str)

    @property
    def running_map(cls) -> Query[str]:
        """Fetches servers running the specified map (e.g. cs_italy)"""
        return Query["str"]("\\map\\", type=str)

    @property
    def ip(cls) -> Query[str]:
        """Fetches servers on the specified IP address, port is optional.

        See Also
        --------
        :meth:`Client.fetch_server` for a Query free version of this.
        """
        return Query["str"]("\\gameaddr\\", type=str)

    @property
    def running(cls) -> Query[Game | int]:
        """Fetches servers running a :class:`.Game` or an :class:`int` app id."""
        return Query["Game | int"]("\\appid\\", type=(Game, int), callback=lambda game: getattr(game, "id", game))

    @property
    def not_running(cls) -> Query[Game | int]:
        """Fetches servers not running a :class:`.Game` or an :class:`int` app id."""
        return Query["Game | int"]("\\nappid\\", type=(Game, int), callback=lambda game: getattr(game, "id", game))

    @property
    def match_tags(cls) -> Query[list[str]]:
        """Fetches servers with all the given tag(s) in :attr:`GameServer.tags`."""
        return Query["list[str]"]("\\gametype\\", type=list, callback=lambda items: f"[{','.join(items)}]")

    @property
    def match_hidden_tags(cls) -> Query[list[str]]:
        """Fetches servers with all the given tag(s) in their 'hidden' tags only applies for :attr:`steam.LFD2`."""
        return Query["list[str]"]("\\gamedata\\", type=list, callback=lambda items: f"[{','.join(items)}]")

    @property
    def match_any_hidden_tags(cls) -> Query[list[str]]:
        """Fetches servers with any of the given tag(s) in their 'hidden' tags only applies for :attr:`steam.LFD2`."""
        return Query["list[str]"]("\\gamedataor\\", type=list, callback=lambda items: f"[{','.join(items)}]")

    @property
    def region(cls) -> Query[GameServerRegion]:
        """Fetches servers in a given region."""
        return Query["GameServerRegion"]("\\region\\", type=GameServerRegion, callback=lambda region: region.value)

    @property
    def all(cls) -> QueryAll:
        """Fetches any servers. Any operations on this will fail."""
        return QueryAll()


class Query(Generic[T_co], metaclass=QueryMeta):
    r"""A :class:`pathlib.Path` like class for constructing Global Master Server queries.

    .. container:: operations

        .. describe:: x == y

            Checks if two queries are equal, order is checked.

        .. describe:: x / y

            Appends y's query to x.

        .. describe:: x | y

            Combines the two queries in ``\nor\[x\y]`` (not or).

        .. describe:: x & y

            Combines the two queries in ``\nand\[x\y]`` (not and).

    Examples
    --------

    Match games running TF2, that are not empty and are using VAC

    .. code-block:: pycon

        >>> Query.running / TF2 / Query.not_empty / Query.secure
        <Query query='\\appid\\440\\empty\\1\\secure\\1'>


    Matches games that are not empty, not full and are not using VAC

    .. code-block:: pycon

        >>> Query.not_empty / Query.not_full | Query.secure
        <Query query='\\empty\\1\\nor\\[\\full\\1\\secure\\1]'>

    Match games where the server name is not "A cool Server" or the server doesn't support alltalk or increased max players

    .. code-block:: pycon

        >>> Query.name_match / "A not cool server" | Query.match_tags / ["alltalk", "increased_maxplayers"]
        <Query query='\\nor\\[\\name_match\\A not cool server\\gametype\\[alltalk,increased_maxplayers]]'>

    Match games where the server is not on linux and the server doesn't have no password (has a password)

    .. code-block:: pycon

        >>> Query.linux & Query.no_password
    """

    # simple specification:
    # - immutable
    # - based on https://developer.valvesoftware.com/wiki/Master_Server_Query_Protocol

    # TODO use __invert__ to do some cool manipulation where possible and generally change the dunders cause they make
    # little sense atm
    __slots__ = ("_raw", "_type", "_callback")
    _raw: tuple[Query[Any], Operator, T_co] | tuple[str]
    _type: type[T_co] | tuple[type[T_co], ...] | None
    _callback: Callable[[T_co], Any] | None

    def __new__(
        cls,
        *raw: Unpack[tuple[Query[Any], Operator, T_co]] | Unpack[tuple[str]],
        type: type[T_co] | tuple[type[T_co], ...] | None = None,
        callback: Callable[[T_co], Any] = lambda x: x,
    ) -> Query[T_co]:
        self = super().__new__(cls)
        self._raw = raw  # type: ignore  # can't tell if this is a pyright bug
        self._type = type
        self._callback = callback
        return self

    def __repr__(self) -> str:
        return f"<Query query={self.query!r}>"

    # it's safe to use a covariant TypeVar here as everything is read-only
    def _process_op(self, other: T_co, op: Operator) -> Q:  # type: ignore
        cls = self.__class__

        if self._type is not None and isinstance(other, self._type):
            return cls(self, op, other)

        if not isinstance(other, Query):
            return NotImplemented

        return cls(self, op, other, type=other._type, callback=other._callback)

    def __truediv__(self, other: T_co) -> Q:  # type: ignore
        return self._process_op(other, Operator.div)

    def __and__(self, other: T_co) -> Q:  # type: ignore
        return self._process_op(other, Operator.nand)

    def __or__(self, other: T_co) -> Q:  # type: ignore
        return self._process_op(other, Operator.nor)

    def __eq__(self, other: object) -> bool:
        return self._raw == other._raw if isinstance(other, Query) else NotImplemented

    @property
    def query(self) -> str:
        """The actual query used for querying Global Master Servers."""

        if len(self._raw) == 1:  # string query
            return self._raw[0]

        # normal
        query_1, op, query_2 = self._raw

        return op.format(
            query_1.query,
            query_2.query if isinstance(query_2, Query) else query_1._callback(query_2),
        )


class ServerPlayer(NamedTuple):
    index: int
    name: str
    score: int
    play_time: timedelta


class GameServer(SteamID):
    """Represents a game server.

    Attributes
    ----------
    name
        The name of the server.
    game
        The game of the server.
    ip
        The ip of the server.
    port
        The port of the server.
    tags
        The tags of the server.
    map
        The map the server is running.
    bot_count
        The number of bots in the server.
    player_count
        The number of players the server.
    max_player_count
        The maximum player count of the server.
    region
        The region the server is in.
    version
        The version of the server.
    """

    __slots__ = (
        "name",
        "game",
        "ip",
        "port",
        "tags",
        "map",
        "bot_count",
        "player_count",
        "max_player_count",
        "region",
        "version",
        "_secure",
        "_dedicated",
        "_loop",
        "_state",
    )

    def __init__(self, state: ConnectionState, server: GetServerListResponseServer):
        super().__init__(server.steamid, type=Type.GameServer)
        self.name = server.name
        self.game = StatefulGame(state, id=server.appid)
        self.ip = server.addr.split(":")[0]  # TODO change to IPv4Address
        self.port = server.gameport
        self.tags = server.gametype.split(",")
        self.map = server.map
        self.bot_count = server.bots
        self.player_count = server.players
        self.max_player_count = server.max_players
        self.region = GameServerRegion.try_value(server.region)
        self.version = server.version

        self._secure = server.secure
        self._dedicated = server.dedicated
        self._state = state

    def __repr__(self) -> str:
        attrs = ("name", "game", "ip", "port", "region", "id", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name

    def is_secure(self) -> bool:
        """Whether the sever is secured, likely with VAC."""
        return self._secure

    def is_dedicated(self) -> bool:
        """Whether the sever is dedicated."""
        return self._dedicated

    async def _query(self, type: EQueryType) -> QueryResponse:
        return await self._state.query_server(int(IPv4Address(self.ip)), self.port, self.game.id, type)

    # async def ping(self):  # FIXME not sure how to expose this
    #     proto = await self._query(EQueryType.Ping)
    #     proto.ping_data

    async def players(self, *, challenge: Literal[-1, 0] = 0) -> list[ServerPlayer]:
        """Fetch a server's players.

        Parameters
        ----------
        challenge
            The challenge for the request default is 0 can also be -1. You may need to change if the server doesn't seem
            to respond.

            .. deprecated:: 0.9.0

                This parameter no longer does anything.

        Returns
        -------
        ServerPlayer is a :class:`typing.NamedTuple` defined as:

        .. source:: steam.ServerPlayer
        """
        if challenge:
            warnings.warn("challenge parameter is deprecated and will be removed in V1", stacklevel=1)

        proto = await self._query(EQueryType.Players)
        return [
            ServerPlayer(
                index=index,
                name=player.name,
                score=player.score,
                play_time=timedelta(seconds=player.time_played),
            )
            for index, player in enumerate(proto.players_data.players)
        ]

    async def rules(self, *, challenge: Literal[-1, 0] = 0) -> dict[str, str]:
        """Fetch a console variables. e.g. ``sv_gravity`` or ``sv_voiceenable``.

        Parameters
        ----------
        challenges
            The challenge for the request default is 0 can also be -1. You may need to change if the server doesn't seem
            to respond.

            .. deprecated:: 0.9.0

                This parameter no longer does anything.
        """

        if challenge:
            warnings.warn("challenge parameter is deprecated and will be removed in V1", stacklevel=1)

        proto = await self._query(EQueryType.Rules)
        return proto.rules_data.rules



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""


from typing_extensions import TypedDict


class FetchedPackageApp(TypedDict):
    id: int
    name: str


class FetchedPackagePrice(TypedDict):
    currency: str
    initial: int
    final: int
    discount_percent: int
    individual: int


class FetchedPackagePlatform(TypedDict):
    windows: bool
    mac: bool
    linux: bool


class FetchedPackageController(TypedDict):
    full_gamepad: bool


class FetchedPackageReleaseDate(TypedDict):
    coming_soon: bool
    date: str


class FetchedPackage(TypedDict):
    packageid: int
    name: str
    page_content: str
    page_image: str
    header_image: str
    small_logo: str
    apps: list[FetchedPackageApp]
    price: FetchedPackagePrice
    platforms: FetchedPackagePlatform
    controller: FetchedPackageController
    release_date: FetchedPackageReleaseDate



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import dis
import sys
import types
import warnings
from collections.abc import Iterator, Sequence
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Generic, NamedTuple, TypeVar, overload

from typing_extensions import Self, TypeAlias

from . import utils
from ._const import URL
from .enums import Language, Result, TradeOfferState
from .errors import ClientException, ConfirmationError, HTTPException
from .game import Game, StatefulGame
from .utils import DateTime

if TYPE_CHECKING:
    from .abc import BaseUser, SteamID
    from .state import ConnectionState
    from .types import trade
    from .user import User


__all__ = (
    "Asset",
    "Item",
    "Inventory",
    "TradeOffer",
    "MovedItem",
    "TradeOfferReceipt",
)

ItemT_co = TypeVar("ItemT_co", bound="Item", covariant=True)


class Asset:
    """Base most version of an item. This class should only be received when Steam fails to find a matching item for
    its class and instance IDs.

    .. container:: operations

        .. describe:: x == y

            Checks if two assets are equal.

        .. describe:: hash(x)

            Returns the hash of an asset.

    Attributes
    -------------
    id
        The assetid of the item.
    amount
        The amount of the same asset there are in the inventory.
    instance_id
        The instanceid of the item.
    class_id
        The classid of the item.
    post_rollback_id
        The assetid of the item after a rollback (cancelled, etc.). ``None`` if not rolled back.
    owner
        The owner of the asset
    """

    __slots__ = (
        "id",
        "amount",
        "class_id",
        "instance_id",
        # "post_rollback_id",
        "owner",
        "_game_cs",
        "_app_id",
        "_context_id",
        "_state",
    )
    REPR_ATTRS = ("id", "class_id", "instance_id", "amount", "owner", "game")  # "post_rollback_id"

    def __init__(self, state: ConnectionState, data: trade.Asset, owner: BaseUser):
        self.id = int(data["assetid"])
        self.amount = int(data["amount"])
        self.instance_id = int(data["instanceid"])
        self.class_id = int(data["classid"])
        # self.post_rollback_id = int(data["rollback_new_assetid"]) if "rollback_new_assetid" in data else None
        self.owner = owner
        self._app_id = int(data["appid"])
        self._context_id = int(data["contextid"])
        self._state = state

    def __repr__(self) -> str:
        cls = self.__class__
        resolved = [f"{attr}={getattr(self, attr, None)!r}" for attr in cls.REPR_ATTRS]
        return f"<{cls.__name__} {' '.join(resolved)}>"

    def __eq__(self, other: Any) -> bool:
        return (
            self.id == other.id and self._app_id == other._app_id and self._context_id == other._context_id
            if isinstance(other, Asset)
            else NotImplemented
        )

    def __hash__(self) -> int:
        return hash((self.id, self._app_id, self._context_id))

    def to_dict(self) -> trade.AssetToDict:
        return {
            "assetid": str(self.id),
            "amount": self.amount,
            "appid": str(self._app_id),
            "contextid": str(self._context_id),
        }

    @utils.cached_slot_property
    def game(self) -> StatefulGame:
        """The game the item is from."""
        return StatefulGame(self._state, id=self._app_id, context_id=self._context_id)

    @property
    def url(self) -> str:
        """The URL for the asset in the owner's inventory.

        e.g. https://steamcommunity.com/profiles/76561198248053954/inventory/#440_2_8526584188
        """
        return f"{URL.COMMUNITY}/profiles/{self.owner.id64}/inventory#{self._app_id}_{self._context_id}_{self.id}"

    @property
    def asset_id(self) -> int:
        """The assetid of the item.

        .. deprecated:: 0.9.0:: Use :attr:`id` instead.
        """
        warnings.warn("asset_id is deprecated, use id instead", DeprecationWarning)
        return self.id


class Item(Asset):
    """Represents an item in a User's inventory.

    .. container:: operations

        .. describe:: x == y

            Checks if two items are equal.

    Attributes
    -------------
    name
        The market_name of the item.
    display_name
        The displayed name of the item. This could be different to :attr:`Item.name` if the item is user re-nameable.
    colour
        The colour of the item.
    descriptions
        The descriptions of the item.
    owner_descriptions
        The descriptions of the item which are visible only to the owner of the item.
    type
        The type of the item.
    tags
        The tags of the item.
    icon_url
        The icon url of the item. Uses the large (184x184 px) image url.
    fraud_warnings
        The fraud warnings for the item.
    actions
        The actions for the item.
    """

    __slots__ = (
        "name",
        "type",
        "tags",
        "colour",
        "icon_url",
        "display_name",
        "descriptions",
        "owner_descriptions",
        "fraud_warnings",
        "actions",
        "owner_actions",
        "market_actions",
        "_is_tradable",
        "_is_marketable",
    )
    REPR_ATTRS = ("name", *Asset.REPR_ATTRS)

    def __init__(self, state: ConnectionState, data: trade.Item, owner: BaseUser):
        super().__init__(state, data, owner)
        self._from_data(data)

    def _from_data(self, data: trade.Item) -> None:
        self.name = data.get("market_name")
        self.display_name = data.get("name")
        self.colour = int(data["name_color"], 16) if "name_color" in data else None
        self.descriptions = data.get("descriptions")
        self.owner_descriptions = data.get("owner_descriptions", [])
        self.type = data.get("type")
        self.tags = data.get("tags")
        self.icon_url = (
            f"https://steamcommunity-a.akamaihd.net/economy/image/{data['icon_url_large']}"
            if "icon_url_large" in data
            else f"https://steamcommunity-a.akamaihd.net/economy/image/{data['icon_url']}"
            if "icon_url" in data
            else None
        )
        self.fraud_warnings = data.get("fraudwarnings", [])
        self.actions = data.get("actions", [])
        self.owner_actions = data.get("owner_actions")
        self.market_actions = data.get("market_actions")
        self._is_tradable = bool(data.get("tradable", False))
        self._is_marketable = bool(data.get("marketable", False))

    def is_tradable(self) -> bool:
        """Whether the item is tradable."""
        return self._is_tradable

    def is_marketable(self) -> bool:
        """Whether the item is marketable."""
        return self._is_marketable


kwargs: dict[str, Any]
if sys.version_info >= (3, 9, 2):  # GenericAlias wasn't subclass-able in 3.9.0
    GenericAlias = types.GenericAlias
    kwargs = {}
else:
    GenericAlias = type(
        types.new_class(
            "",
            (Generic[TypeVar("T")],),  # type: ignore
        )[int]
    )
    kwargs = {"_root": True}


class InventoryGenericAlias(GenericAlias, **kwargs):
    __alias_name__: str
    __args__: tuple[type[ItemT_co]]  # type: ignore

    def __repr__(self) -> str:
        return f"{self.__origin__.__module__}.{object.__getattribute__(self, '__alias_name__')}"

    def __call__(self, *args: Any, **kwargs: Any) -> object:
        # this is done cause we need __orig_class__ in __init__
        result = self.__origin__.__new__(self.__origin__, *args, **kwargs)
        try:
            result.__orig_class__ = self
        except AttributeError:
            pass
        result.__init__(*args, **kwargs)
        return result

    def __mro_entries__(self, bases: tuple[type, ...]) -> tuple[type]:
        # if we are subclassing we should return a new class that already has __orig_class__

        class BaseInventory(*super().__mro_entries__(bases)):  # type: ignore
            __slots__ = ()
            __orig_class__ = self

        return (BaseInventory,)


class BaseInventory(Generic[ItemT_co]):
    """Base for all inventories."""

    __slots__ = (
        "game",
        "items",
        "owner",
        "_language",
        "_state",
        "__orig_class__",  # undocumented typing internals more shim to make extensions work
    )
    __orig_class__: InventoryGenericAlias

    def __init__(
        self, state: ConnectionState, data: trade.Inventory, owner: BaseUser, game: Game, language: Language | None
    ):
        self._state = state
        self.owner = owner
        self.game = StatefulGame(state, id=game.id, name=game.name)
        self._language = language
        self._update(data)

    def __repr__(self) -> str:
        attrs = ("owner", "game")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__orig_class__} {' '.join(resolved)}>"

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Iterator[ItemT_co]:
        return iter(self.items)

    def __contains__(self, item: Asset) -> bool:
        if not isinstance(item, Asset):
            raise TypeError(
                f"unsupported operand type(s) for 'in': {item.__class__.__qualname__!r} and {self.__orig_class__!r}"
            )
        return item in self.items

    if not TYPE_CHECKING:

        def __class_getitem__(cls, params: tuple[type[ItemT_co]]) -> InventoryGenericAlias:
            # this is more stuff that's needed to make the TypeAliases for extension modules work as we need the
            # assigned name

            frame = sys._getframe(1)
            generic_alias = InventoryGenericAlias(cls, params)

            on_line = False
            return_next = False
            for instruction in dis.get_instructions(frame.f_code):
                if return_next and instruction.opname == "STORE_NAME":
                    break
                elif instruction.starts_line == frame.f_lineno:
                    on_line = True
                elif on_line and instruction.opname == "BINARY_SUBSCR":
                    return_next = True
            else:
                return generic_alias

            object.__setattr__(generic_alias, "__alias_name__", instruction.argval)
            return generic_alias

    def _update(self, data: trade.Inventory) -> None:
        items = []
        (ItemClass,) = self.__orig_class__.__args__
        for asset in data.get("assets", ()):
            for item in data["descriptions"]:
                if item["instanceid"] == asset["instanceid"] and item["classid"] == asset["classid"]:
                    item.update(asset)  # type: ignore  # maybe this will work if types.IntersectionType happens
                    items.append(ItemClass(self._state, data=item, owner=self.owner))  # type: ignore
                    break
            else:
                items.append(Asset(self._state, data=asset, owner=self.owner))
        self.items: Sequence[ItemT_co] = items

    async def update(self) -> None:
        """Re-fetches the inventory and updates it inplace."""
        if self.owner == self._state.user:
            data = await self._state.http.get_client_user_inventory(self.game.id, self.game.context_id, self._language)
        else:
            data = await self._state.http.get_user_inventory(
                self.owner.id64, self.game.id, self.game.context_id, self._language
            )
        self._update(data)

    def filter_items(self, *names: str, limit: int | None = None) -> list[ItemT_co]:
        """A helper function that filters items by name from the inventory.

        .. deprecated:: 0.9

            Use a list comprehension instead of this.

        Parameters
        ------------
        names
            The names of the items to filter for.
        limit
            The maximum amount of items to return. Checks from the front of the items.

        Raises
        -------
        :exc:`ValueError`
            You passed a limit and multiple item names.

        Returns
        ---------
        The matching items.
        """
        if len(names) > 1 and limit:
            raise ValueError("Cannot pass a limit with multiple items")
        items = [item for item in self if item.name in names]
        return items if limit is None else items[:limit]

    def get_item(self, name: str) -> ItemT_co | None:
        """A helper function that gets an item or ``None`` if no matching item is found by name from the inventory.

        .. deprecated:: 0.9

            Use :func:`steam.utils.get` instead of this.

        Parameters
        ----------
        name
            The item to get from the inventory.
        """
        item = self.filter_items(name, limit=1)
        return item[0] if item else None


Inventory: TypeAlias = BaseInventory[Item]  # necessitated by TypeVar not currently supporting defaults
"""Represents a User's inventory.

.. container:: operations

    .. describe:: len(x)

        Returns how many items are in the inventory.

    .. describe:: iter(x)

        Iterates over the inventory's items.

    .. describe:: y in x

        Determines if an item is in the inventory based off of its :attr:`Asset.id`.


Attributes
----------
items
    A list of the inventory's items.
owner
    The owner of the inventory.
game
    The game the inventory the game belongs to.
"""


class TradeOfferReceipt(NamedTuple):
    sent: list[MovedItem]
    received: list[MovedItem]


class MovedItem(Item):
    """Represents an item that has moved from one inventory to another.

    Attributes
    ----------
    new_id
        The new_assetid field, this is the asset ID of the item in the partners inventory.
    new_context_id
        The new_contextid field.
    """

    __slots__ = (
        "new_id",
        "new_context_id",
    )
    REPR_ATTRS = (*Item.REPR_ATTRS, "new_id", "new_context_id")
    new_id: int
    new_context_id: int

    def _from_data(self, data: trade.TradeOfferReceiptItem):
        super()._from_data(data)
        self.new_id = int(data["new_assetid"])
        self.new_context_id = int(data["new_contextid"])

    @property
    def new_asset_id(self) -> int:
        """The new asset ID of the item in the partner's inventory."""
        warnings.warn("new_asset_id is deprecated, use new_id instead", DeprecationWarning)
        return self.new_id


class TradeOffer:
    """Represents a trade offer from/to send to a User.
    This can also be used in :meth:`steam.User.send`.

    Parameters
    ----------
    item_to_send
        The item to send with the trade offer. Mutually exclusive to ``items_to_send``.
    item_to_receive
        The item to receive with the trade offer. Mutually exclusive to ``items_to_receive``.
    items_to_send
        The items you are sending to the other user. Mutually exclusive to ``item_to_send``.
    items_to_receive
        The items you are receiving from the other user. Mutually exclusive to ``item_to_receive``.
    token
        The trade token used to send trades to users who aren't on the ClientUser's friend's list.
    message
         The offer message to send with the trade.

    Attributes
    ----------
    partner
        The trade offer partner. This should only ever be a :class:`~steam.SteamID` if the partner's profile is private.
    items_to_send
        A list of items to send to the partner.
    items_to_receive
        A list of items to receive from the partner.
    state
        The offer state of the trade for the possible types see :class:`~steam.TradeOfferState`.
    message
        The message included with the trade offer.
    id
        The trade's offer ID.
    created_at
        The time at which the trade was created.
    updated_at
        The time at which the trade was last updated.
    expires
        The time at which the trade automatically expires.
    escrow
        The time at which the escrow will end. Can be ``None`` if there is no escrow on the trade.

        Warning
        -------
        This isn't likely to be accurate, use :meth:`User.escrow` instead if possible.
    """

    __slots__ = (
        "id",
        "_id",
        "state",
        "escrow",
        "partner",
        "message",
        "token",
        "expires",
        "updated_at",
        "created_at",
        "items_to_send",
        "items_to_receive",
        "_has_been_sent",
        "_state",
        "_is_our_offer",
    )

    id: int
    partner: User | SteamID

    @overload
    def __init__(
        self,
        *,
        token: str | None = ...,
        message: str | None = ...,
        item_to_send: Asset = ...,
        item_to_receive: Asset = ...,
    ):
        ...

    @overload
    def __init__(
        self,
        *,
        token: str | None = ...,
        message: str | None = ...,
        items_to_send: Sequence[Asset],
        items_to_receive: Sequence[Asset],
    ):
        ...

    def __init__(
        self,
        *,
        message: str | None = None,
        token: str | None = None,
        item_to_send: Asset | None = None,
        item_to_receive: Asset | None = None,
        items_to_send: Sequence[Asset] | None = None,
        items_to_receive: Sequence[Asset] | None = None,
    ):
        self.items_to_receive: Sequence[Asset] = items_to_receive or ([item_to_receive] if item_to_receive else [])
        self.items_to_send: Sequence[Asset] = items_to_send or ([item_to_send] if item_to_send else [])
        self.message: str | None = message or None
        self.token: str | None = token
        self.updated_at: datetime | None = None
        self.created_at: datetime | None = None
        self.escrow: timedelta | None = None
        self.state = TradeOfferState.Invalid
        self._id: int | None = None
        self._has_been_sent = False

    @classmethod
    def _from_api(cls, state: ConnectionState, data: trade.TradeOffer, partner: User | SteamID | None = None) -> Self:
        trade = cls()
        trade._has_been_sent = True
        trade._state = state
        if partner is None:
            from .abc import SteamID

            trade.partner = SteamID(data["accountid_other"])
        else:
            trade.partner = partner
        trade._update(data)
        return trade

    @classmethod
    def _from_history(cls, state: ConnectionState, data: trade.TradeOfferHistoryTrade) -> Self:
        received: list[trade.TradeOfferReceiptItem] = data.get("assets_received", [])  # type: ignore
        sent: list[trade.TradeOfferReceiptItem] = data.get("assets_given", [])  # type: ignore
        from .abc import SteamID

        partner = SteamID(data["steamid_other"])
        trade = cls(
            items_to_receive=[MovedItem(state, item, partner) for item in received],
            items_to_send=[MovedItem(state, item, state.user) for item in sent],
        )
        trade._state = state
        trade._id = int(data["tradeid"])
        trade.partner = partner
        trade.created_at = DateTime.from_timestamp(data["time_init"])
        trade.state = TradeOfferState.try_value(data["status"])

        return trade

    def _update_from_send(
        self, state: ConnectionState, data: trade.TradeOfferCreateResponse, partner: User, active: bool = True
    ) -> None:
        self.id = int(data["tradeofferid"])
        self._state = state
        self.partner = partner
        self.state = TradeOfferState.Active if active else TradeOfferState.ConfirmationNeed
        self.created_at = DateTime.now()
        self._is_our_offer = True

    def __repr__(self) -> str:
        attrs = ("id", "state", "partner")
        resolved = [f"{attr}={getattr(self, attr, None)!r}" for attr in attrs]
        return f"<TradeOffer {' '.join(resolved)}>"

    def _update(self, data: trade.TradeOffer) -> None:
        self.message = data.get("message") or None
        self.id = int(data["tradeofferid"])
        self._id = int(data["tradeid"]) if "tradeid" in data else None
        expires = data.get("expiration_time")
        escrow = data.get("escrow_end_date")
        updated_at = data.get("time_updated")
        created_at = data.get("time_created")
        self.expires = DateTime.from_timestamp(expires) if expires else None
        self.escrow = DateTime.from_timestamp(escrow) - DateTime.now() if escrow else None
        self.updated_at = DateTime.from_timestamp(updated_at) if updated_at else None
        self.created_at = DateTime.from_timestamp(created_at) if created_at else None
        self.state = TradeOfferState.try_value(data.get("trade_offer_state", 1))
        self.items_to_send = [
            Item(self._state, data=item, owner=self.partner) for item in data.get("items_to_give", [])
        ]
        self.items_to_receive = [
            Item(self._state, data=item, owner=self.partner) for item in data.get("items_to_receive", [])
        ]
        self._is_our_offer = data.get("is_our_offer", False)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TradeOffer):
            return NotImplemented
        if self._has_been_sent and other._has_been_sent:
            return self.id == other.id
        elif not (self._has_been_sent and other._has_been_sent):
            return self.items_to_send == other.items_to_send and self.items_to_receive == other.items_to_receive
        return NotImplemented

    async def confirm(self) -> None:
        """Confirms the trade offer.
        This rarely needs to be called as the client handles most of these.

        Raises
        ------
        steam.ClientException
            The trade is not active.
        steam.ConfirmationError
            No matching confirmation could not be found.
        """
        self._check_active()
        if self.is_gift():
            return  # no point trying to confirm it
        if not await self._state.fetch_and_confirm_confirmation(self.id):
            raise ConfirmationError("No matching confirmation could be found for this trade")
        self._state._confirmations.pop(self.id, None)

    async def accept(self) -> None:
        """Accepts the trade offer.

        Note
        ----
        This also calls :meth:`confirm` (if necessary) so you don't have to.

        Raises
        ------
        steam.ClientException
            The trade is either not active, already accepted or not from the ClientUser.
        steam.ConfirmationError
            No matching confirmation could not be found.
        """
        if self.state == TradeOfferState.Accepted:
            raise ClientException("This trade has already been accepted")
        if self.is_our_offer():
            raise ClientException("You cannot accept an offer the ClientUser has made")
        self._check_active()
        assert self.partner is not None
        resp = await self._state.http.accept_user_trade(self.partner.id64, self.id)
        if resp.get("needs_mobile_confirmation", False):
            for tries in range(5):
                try:
                    return await self.confirm()
                except ConfirmationError:
                    break
                except ClientException:
                    if tries == 4:
                        raise ClientException("Failed to accept trade offer") from None
                    await asyncio.sleep(tries * 2)

    async def decline(self) -> None:
        """Declines the trade offer.

        Raises
        ------
        :exc:`~steam.ClientException`
            The trade is either not active, already declined or not from the ClientUser.
        """
        if self.state == TradeOfferState.Declined:
            raise ClientException("This trade has already been declined")
        if self.is_our_offer():
            raise ClientException("You cannot decline an offer the ClientUser has made")
        self._check_active()
        await self._state.http.decline_user_trade(self.id)

    async def cancel(self) -> None:
        """Cancels the trade offer.

        Raises
        ------
        :exc:`~steam.ClientException`
            The trade is either not active or already cancelled.
        """
        if self.state == TradeOfferState.Canceled:
            raise ClientException("This trade has already been cancelled")
        self._check_active()
        await self._state.http.cancel_user_trade(self.id)

    async def receipt(self) -> TradeOfferReceipt:
        """Get the receipt for a trade offer and the updated asset ids for the trade.

        Returns
        -------
        A trade receipt.

        .. source:: steam.TradeOfferReceipt
        """
        if self._id is None:
            raise ValueError("Cannot fetch the receipt for a trade not accepted")

        resp = await self._state.http.get_trade_receipt(self._id)
        data = resp["response"]
        trade = data["trades"][0]
        descriptions = data["descriptions"]
        assert self.partner is not None

        received: list[MovedItem] = []
        for asset in trade.get("assets_received", ()):
            for item in descriptions:
                if item["instanceid"] == asset["instanceid"] and item["classid"] == asset["classid"]:
                    item.update(asset)
                    received.append(MovedItem(self._state, data=item, owner=self.partner))  # type: ignore

        sent: list[MovedItem] = []
        for asset in trade.get("assets_given", ()):
            for item in descriptions:
                if item["instanceid"] == asset["instanceid"] and item["classid"] == asset["classid"]:
                    item.update(asset)
                    sent.append(MovedItem(self._state, data=item, owner=self._state.user))

        return TradeOfferReceipt(sent=sent, received=received)

    async def counter(self, trade: TradeOffer) -> None:
        """Counter a trade offer from an :class:`User`.

        Parameters
        -----------
        trade
            The trade offer to counter with.

        Raises
        ------
        :exc:`~steam.ClientException`
            The trade from the ClientUser or it isn't active.
        """
        self._check_active()
        if self.is_our_offer():
            raise ClientException("You cannot counter an offer the ClientUser has made")

        to_send = [item.to_dict() for item in trade.items_to_send]
        to_receive = [item.to_dict() for item in trade.items_to_receive]
        assert self.partner is not None
        resp = await self._state.http.send_trade_offer(
            self.partner, to_send, to_receive, trade.token, trade.message or "", trade_id=self.id
        )
        if resp.get("needs_mobile_confirmation", False):
            await self._state.fetch_and_confirm_confirmation(int(resp["tradeofferid"]))

    @property
    def url(self) -> str:
        """The URL of the trade offer."""
        return str(URL.COMMUNITY / f"tradeoffer/{self.id}")

    def is_gift(self) -> bool:
        """Helper method that checks if an offer is a gift to the :class:`~steam.ClientUser`"""
        return bool(self.items_to_receive and not self.items_to_send)

    def is_our_offer(self) -> bool:
        """Whether the offer was created by the :class:`~steam.ClientUser`."""
        return self._is_our_offer

    def _check_active(self) -> None:
        if self.state not in (TradeOfferState.Active, TradeOfferState.ConfirmationNeed) or not self._has_been_sent:
            raise ClientException("This trade is not active")



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import itertools
import math
import re
from collections.abc import AsyncGenerator, Callable, Coroutine, Iterable
from datetime import datetime
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from bs4 import BeautifulSoup
from typing_extensions import ClassVar, Self, TypeAlias
from yarl import URL as URL_

from . import utils
from ._const import HTML_PARSER, URL
from .enums import EventType, Language, PublishedFileQueryFileType, PublishedFileRevision, PublishedFileType
from .utils import DateTime

if TYPE_CHECKING:
    from .abc import Authors, BaseUser, Commentable, Message
    from .channel import DMChannel, UserMessage
    from .chat import Chat, ChatMessage
    from .clan import Clan
    from .comment import Comment
    from .event import Announcement, Event
    from .game import Game, StatefulGame
    from .group import Group
    from .manifest import Manifest
    from .protobufs import friend_messages
    from .published_file import PublishedFile
    from .review import Review
    from .state import ConnectionState
    from .trade import TradeOffer
    from .user import User

T = TypeVar("T")
TT = TypeVar("TT")
CommentableT = TypeVar("CommentableT", bound="Commentable")
M = TypeVar("M", bound="Message", covariant=True)

MaybeCoro: TypeAlias = "Callable[[T], bool | Coroutine[Any, Any, bool]]"
UNIX_EPOCH = DateTime.from_timestamp(0)


class AsyncIterator(Generic[T]):
    """A class from which async iterators (see :pep:`525`) can ben easily derived.

    .. container:: operations

        .. describe:: async for y in x

            Iterates over the contents of the async iterator.

    Attributes
    ----------
    before
        When to find objects before.
    after
        When to find objects after.
    limit
        The maximum number of elements to be yielded.
    """

    def __init__(
        self,
        state: ConnectionState,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ):
        self._state = state
        self.before = before or DateTime.now()
        self.after = after or UNIX_EPOCH
        self._is_filled = False
        self.limit = limit
        self._fill = self.fill()
        self._seen = 0

    async def _fill_queue_users(
        self,
        iterable: Iterable[T],
        attributes: tuple[str, ...] = ("author",),
        *,
        user_attribute_name: str | None = None,
    ) -> None:
        user_attribute_name = user_attribute_name or attributes[0]
        users = await self._state._maybe_users(getattr(element, user_attribute_name) for element in iterable)
        for user, element in itertools.product(
            users,
            iterable,
        ):
            for attribute in attributes:
                if getattr(element, attribute, None) == user:
                    setattr(element, attribute, user)

    async def get(self, **attrs: Any) -> T | None:
        """A helper function which is similar to :func:`~steam.utils.get` except it runs over the async iterator.

        This is roughly equivalent to:

        .. code-block:: python3


            elements = await async_iterator.flatten()
            element = steam.utils.get(elements, name="Item")

        Example
        -------
        Getting the last comment from a user named 'Dave' or None:

        .. code-block:: python3

            comment = await user.comments().get(author__name="Dave")

        Parameters
        ----------
        attrs
            Keyword arguments that denote attributes to match.

        Returns
        -------
        The first element from the ``iterable`` which matches all the traits passed in ``attrs`` or ``None`` if no
        matching element was found.
        """

        def predicate(elem: T) -> bool:
            for attr, val in attrs.items():
                nested = attr.split("__")
                obj = elem
                for attribute in nested:
                    obj = getattr(obj, attribute)

                if obj != val:
                    return False
            return True

        return await self.find(predicate)

    async def find(self, predicate: MaybeCoro[T]) -> T | None:
        """A helper function which is similar to :func:`~steam.utils.find` except it runs over the async iterator.
        However, unlike :func:`~steam.utils.find`, the predicate provided can be a |coroutine_link|_.

        This is roughly equivalent to:

        .. code-block:: python3

            elements = await async_iterator.flatten()
            element = steam.utils.find(elements, lambda e: e.name == "Item")

        Example
        -------
        Getting the last trade with a message or None:

        .. code-block:: python3

            def predicate(trade: steam.TradeOffer) -> bool:
                return trade.message is not None


            trade = await client.trade_history().find(predicate)

        Parameters
        ----------
        predicate
            A callable/coroutine that returns a boolean.

        Returns
        -------
        The first element from the iterator for which the ``predicate`` returns ``True`` or ``None`` if no matching
        element was found.
        """
        async for elem in self:
            ret = await utils.maybe_coroutine(predicate, elem)
            if ret:
                return elem

    async def flatten(self) -> list[T]:
        """A helper function that iterates over the :class:`AsyncIterator` returning a list of all the elements in the
        iterator.

        This is equivalent to:

        .. code-block:: python3

            elements = [element async for element in async_iterator]
        """
        return [element async for element in self]

    def filter(self, predicate: Callable[[T], bool]) -> FilteredIterator[T]:
        """Filter members of the async iterator according to a predicate. This function acts similarly to :func:`filter`.

        Examples
        --------
        .. code-block:: python3

            for dave in async_iterator.filter(lambda x: x.name == "Dave"):
                ...  # the element now has to have a name of Dave.

        Parameters
        ----------
        predicate
            The predicate to filter elements through.
        """
        return FilteredIterator(predicate, self)

    def map(self, func: Callable[[TT], Any]) -> MappedIterator[T, TT]:
        """Map the elements of the async iterator through a function. This function acts similarly to :func:`map`.

        Examples
        --------
        .. code-block:: python3

            for name in async_iterator.map(lambda x: x.name):
                ...  # name is now the iterators element's name.

        Parameters
        ----------
        func
            The function to map the elements through.
        """
        return MappedIterator(func, self)

    def __aiter__(self) -> Self:
        return self

    def __anext__(self) -> Coroutine[None, None, T]:
        self._seen += 1
        if self.limit and self._seen > self.limit:
            raise StopAsyncIteration
        return self._fill.__anext__()  # type: ignore  # this is typed wrong

    async def next(self) -> T:
        """Advances the iterator by one, if possible.

        Raises
        ------
        :exc:`StopAsyncIteration`
            There are no more elements in the iterator.
        """
        return await self.__anext__()

    async def fill(self) -> AsyncGenerator[T, None]:
        raise NotImplementedError
        yield


class FilteredIterator(AsyncIterator[T]):
    def __init__(self, predicate: MaybeCoro[T], async_iterator: AsyncIterator[T]):
        self.predicate = predicate
        self.iterator = async_iterator

    async def next(self) -> T:
        while True:
            item = await self.iterator.next()
            if await utils.maybe_coroutine(self.predicate, item):
                return item


class MappedIterator(AsyncIterator[TT], Generic[T, TT]):
    def __init__(self, map_func: Callable[[Any], TT | Coroutine[Any, Any, TT]], async_iterator: AsyncIterator[T]):
        self.map_func = map_func
        self.iterator = async_iterator

    async def next(self) -> TT:
        item = await self.iterator.next()
        return await utils.maybe_coroutine(self.map_func, item)


class CommentsIterator(AsyncIterator["Comment[CommentableT]"]):
    def __init__(
        self,
        owner: CommentableT,
        oldest_first: bool,
        state: ConnectionState,
        limit: int | None,
        before: datetime | None,
        after: datetime | None,
    ):
        self.owner = owner
        self.oldest_first = oldest_first
        super().__init__(state, limit, before, after)

    async def fill(self) -> AsyncGenerator["Comment[CommentableT]", None]:
        from .comment import Comment
        from .reaction import AwardReaction

        after = self.after
        before = self.before
        count: int | None = None
        total_count = 0

        async def get_comments(chunk: int) -> list[Comment[CommentableT]]:
            nonlocal after, before, count, total_count

            starting_from = total_count - count if total_count and count else 0
            proto = await self._state.fetch_comments(self.owner, chunk, starting_from, self.oldest_first)

            if count is None:
                total_count = count = proto.total_count

            comments: list[Comment[CommentableT]] = []
            comment = None
            for comment in proto.comments:
                comment = Comment(
                    self._state,
                    id=comment.id,
                    content=comment.content,
                    created_at=DateTime.from_timestamp(comment.timestamp),
                    reactions=[AwardReaction(self._state, reaction) for reaction in comment.reactions],
                    author=comment.author_id64,  # type: ignore
                    owner=self.owner,
                )
                if self.after < comment.created_at < self.before:
                    comments.append(comment)
                else:
                    break

            await self._fill_queue_users(comments)
            count -= len(comments)
            return comments

        for comment in await get_comments(min(self.limit or 100, 100)):
            yield comment

        assert count is not None
        while count > 0:
            for comment in await get_comments(min(count, 100)):
                yield comment


class TradesIterator(AsyncIterator["TradeOffer"]):
    def __init__(
        self,
        state: ConnectionState,
        limit: int | None,
        before: datetime | None,
        after: datetime | None,
        language: Language | None,
    ):
        super().__init__(state, limit, before, after)
        self.language = language

    async def fill(self) -> AsyncGenerator[TradeOffer, None]:
        from .trade import TradeOffer

        total = 100
        previous_time = 0
        after_timestamp = self.after.timestamp()
        before_timestamp = self.before.timestamp()

        async def get_trades(page: int = 100) -> list[TradeOffer]:
            nonlocal total, previous_time
            resp = await self._state.http.get_trade_history(page, previous_time, self.language)
            data = resp["response"]
            if total is None:
                total = data.get("total_trades", 0)
            if not total:
                return []

            trades: list[TradeOffer] = []
            descriptions = data.get("descriptions", ())
            trade = None
            for trade in data.get("trades", []):
                if not after_timestamp < trade["time_init"] < before_timestamp:
                    break
                for item in descriptions:
                    for asset in trade.get("assets_received", []):
                        if item["classid"] == asset["classid"] and item["instanceid"] == asset["instanceid"]:
                            asset.update(item)
                    for asset in trade.get("assets_given", []):
                        if item["classid"] == asset["classid"] and item["instanceid"] == asset["instanceid"]:
                            asset.update(item)

                trades.append(TradeOffer._from_history(state=self._state, data=trade))

            assert trade is not None
            previous_time = trade["time_init"]
            await self._fill_queue_users(trades, ("partner",))
            return trades

        for trade in await get_trades():
            for item in trade.items_to_receive:
                item.owner = trade.partner
            yield trade

        if total < 100:
            for page in range(200, math.ceil((total + 100) / 100) * 100, 100):
                for trade in await get_trades(page):
                    for item in trade.items_to_receive:
                        item.owner = trade.partner
                    yield trade


class DMChannelHistoryIterator(AsyncIterator["UserMessage"]):
    def __init__(
        self,
        channel: DMChannel,
        state: ConnectionState,
        limit: int | None,
        before: datetime | None,
        after: datetime | None,
    ):
        super().__init__(state, limit, before, after)
        self.channel = channel
        self.participant = channel.participant

    async def fill(self) -> AsyncGenerator[UserMessage, None]:
        from .message import Message, UserMessage
        from .reaction import Emoticon, MessageReaction, Sticker

        after_timestamp = int(self.after.timestamp())
        before_timestamp = int(self.before.timestamp())

        last_message_timestamp = before_timestamp
        ordinal = 0

        while True:
            resp = await self._state.fetch_user_history(
                self.participant.id64, start=after_timestamp, last=last_message_timestamp, start_ordinal=ordinal
            )

            message: friend_messages.GetRecentMessagesResponseFriendMessage | None = None

            for message in resp.messages:
                new_message = UserMessage.__new__(UserMessage)
                new_message.created_at = DateTime.from_timestamp(message.timestamp)
                if not self.after < new_message.created_at < self.before:
                    return

                Message.__init__(new_message, channel=self.channel, proto=message)
                new_message.author = self.participant if message.accountid == self.participant.id else self._state.user
                emoticon_reactions = [
                    MessageReaction(
                        self._state,
                        new_message,
                        Emoticon(self._state, r.reaction),
                        None,
                        self.participant if reactor == self.participant.id else self._state.user,
                    )
                    for r in message.reactions
                    if r.reaction_type == 1
                    for reactor in r.reactors
                ]
                sticker_reactions = [
                    MessageReaction(
                        self._state,
                        new_message,
                        None,
                        Sticker(self._state, r.reaction),
                        self.participant if reactor == self.participant.id else self._state.user,
                    )
                    for r in message.reactions
                    if r.reaction_type == 2
                    for reactor in r.reactors
                ]
                new_message.reactions = emoticon_reactions + sticker_reactions

                yield new_message

            if message is None:
                return

            last_message_timestamp = message.timestamp
            ordinal = message.ordinal

            if not resp.more_available:
                return


ChatMessageT = TypeVar("ChatMessageT", bound="ChatMessage", covariant=True)
ChatT = TypeVar("ChatT", bound="Chat[Any]", covariant=True)


class ChatHistoryIterator(AsyncIterator[ChatMessageT], Generic[ChatMessageT, ChatT]):
    def __init__(
        self,
        channel: ChatT,
        state: ConnectionState,
        limit: int | None,
        before: datetime | None,
        after: datetime | None,
    ):
        super().__init__(state, limit, before, after)
        self.channel = channel
        self.group: Group | Clan = channel.group or channel.clan  # type: ignore

    async def fill(self) -> AsyncGenerator[ChatMessageT, None]:
        from .abc import SteamID
        from .message import Message
        from .reaction import Emoticon, PartialMessageReaction, Sticker

        after_timestamp = int(self.after.timestamp())
        before_timestamp = int(self.before.timestamp())
        last_message_timestamp = before_timestamp
        last_ordinal: int = getattr(self.channel.last_message, "ordinal", 0)
        message_cls: type[ChatMessageT] = self.channel._type_args[0]

        while True:
            resp = await self._state.fetch_group_history(
                *self.channel._location, start=after_timestamp, last=last_message_timestamp, last_ordinal=last_ordinal
            )
            message = None
            messages: list[ChatMessageT] = []

            for message in resp.messages:
                new_message = message_cls.__new__(message_cls)
                Message.__init__(new_message, channel=self.channel, proto=message)
                new_message.created_at = DateTime.from_timestamp(message.server_timestamp)
                if not self.after < new_message.created_at < self.before:
                    return

                new_message.author = SteamID(message.sender)
                emoticon_reactions = [
                    PartialMessageReaction(
                        self._state,
                        new_message,
                        Emoticon(self._state, r.reaction),
                        None,
                    )
                    for r in message.reactions
                    if r.reaction_type == 1
                ]
                sticker_reactions = [
                    PartialMessageReaction(
                        self._state,
                        new_message,
                        None,
                        Sticker(self._state, r.reaction),
                    )
                    for r in message.reactions
                    if r.reaction_type == 2
                ]
                new_message.partial_reactions = emoticon_reactions + sticker_reactions

                messages.append(new_message)

            if message is None:
                return

            await self._fill_queue_users(messages)
            for message in messages:
                yield message

            last_message_timestamp = message.server_timestamp
            last_ordinal = message.ordinal

            if not resp.more_available:
                return


class GameReviewsIterator(AsyncIterator["Review"]):
    def __init__(
        self,
        state: ConnectionState,
        game: StatefulGame,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ):  # TODO add support for the other params and make this more efficient
        super().__init__(state, limit, before, after)
        self.game = game

    async def fill(self) -> AsyncGenerator[Review, None]:
        from .review import Review, ReviewGame

        cursor = "*"
        while True:
            data = await self._state.http.get_reviews(self.game.id, "all", "all", "all", cursor)
            if cursor == "*":
                self.game = ReviewGame(self._state, self.game.id, data["query_summary"]["review_score"])
            assert isinstance(self.game, ReviewGame)
            cursor = data["cursor"]
            reviews = data["reviews"]

            for review, user in zip(
                reviews,
                await self._state.fetch_users(int(review["author"]["steamid"]) for review in reviews),
            ):
                if user is None:
                    continue
                review = Review._from_data(self._state, review, self.game, user)
                if not self.after < review.created_at < self.before:
                    return

                yield review


class UserReviewsIterator(AsyncIterator["Review"]):
    def __init__(
        self,
        state: ConnectionState,
        user: BaseUser,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ):
        super().__init__(state, limit, before, after)
        self.user = user

    async def fill(self) -> AsyncGenerator[Review, None]:
        from .review import Review

        pages = 1

        # ideally I'd like to find an actual api for these
        async def get_app_ids(page_number: int = 1) -> list[int]:
            nonlocal pages
            page = await self._state.http.get(
                URL.COMMUNITY / f"profiles/{self.user.id64}/recommended", params={"p": page_number}
            )
            soup = BeautifulSoup(page, HTML_PARSER)
            if pages == 1:
                *_, pages = [1] + [
                    int(a["href"][len("?p=") :]) for a in soup.find_all("a", class_="pagelink")
                ]  # str.removeprefix
            return [
                int(URL_(review.find("div", class_="leftcol").a["href"]).parts[-1])
                for review in soup.find_all("div", class_="review_box_content")
            ]

        for review_ in await self._state.fetch_user_review(self.user.id64, await get_app_ids()):
            review = Review._from_proto(self._state, review_, self.user)
            if not self.after < review.created_at < self.before:
                return

            yield review

        for page in range(2, pages + 1):
            for review_ in await self._state.fetch_user_review(self.user.id64, await get_app_ids(page)):
                review = Review._from_proto(self._state, review_, self.user)
                if not self.after < review.created_at < self.before:
                    return

                yield review


class _EventIterator(AsyncIterator[T]):
    ID_PARSE_REGEX: ClassVar[re.Pattern[str]]

    def __init__(
        self, clan: Clan, state: ConnectionState, limit: int | None, before: datetime | None, after: datetime | None
    ):
        super().__init__(state, limit, before, after)
        self.clan = clan

    async def fill(self) -> AsyncGenerator[T, None]:
        cls = self.__class__
        rss = await self._state.http.get_clan_rss(
            self.clan.id64
        )  # TODO make this use the calendar? does that work for announcements

        soup = BeautifulSoup(rss, HTML_PARSER)

        ids = []
        for url in soup.find_all("guid"):
            match = cls.ID_PARSE_REGEX.findall(url.text)
            if match:
                ids.append(int(match[0]))

        if not ids:
            return

        from . import event

        event_cls: type[Announcement | Event[EventType]] = getattr(
            event, cls.__orig_bases__[0].__args__[0].__forward_arg__  # type: ignore
        )
        events = []
        for event_ in await self.get_events(ids):
            event = event_cls(self._state, self.clan, event_)
            if not self.after < event.starts_at < self.before:
                return
            events.append(event)

        await self._fill_queue_users(events, ("author", "last_edited_by", "approved_by"))

        for event in events:
            yield event

    async def get_events(self, ids: list[int]) -> list[dict[str, Any]]:
        raise NotImplementedError


class EventIterator(_EventIterator["Event[EventType]"]):
    ID_PARSE_REGEX = re.compile(r"events/+(\d+)")

    async def get_events(self, ids: list[int]) -> list[dict[str, Any]]:
        data = await self._state.http.get_clan_events(self.clan.id, ids)
        return data["events"]


class AnnouncementsIterator(_EventIterator["Announcement"]):
    ID_PARSE_REGEX = re.compile(r"announcements/detail/(\d+)")

    async def get_events(self, ids: list[int]) -> list[dict[str, Any]]:
        announcements = await asyncio.gather(*(self._state.http.get_clan_announcement(self.clan.id, id) for id in ids))
        events = []
        for announcement in announcements:
            events += announcement["events"]
        return events


class ManifestIterator(AsyncIterator["Manifest"]):
    def __init__(
        self,
        state: ConnectionState,
        limit: int | None,
        before: datetime | None,
        after: datetime | None,
        game: StatefulGame,
        branch: str,
        password: str | None,
        password_hash: str,
    ):
        super().__init__(state, limit, before, after)
        self.game = game
        self.branch = branch
        self.password = password
        self.password_hash = password_hash

    async def fill(self) -> AsyncGenerator[Manifest, None]:
        manifest_coros = await self._state.fetch_manifests(
            self.game.id, self.branch, self.password, self.limit, self.password_hash
        )
        for chunk in utils.chunk(manifest_coros, 100):
            for manifest in await asyncio.gather(*chunk):
                if self.after < manifest.created_at < self.before:
                    yield manifest


class UserPublishedFilesIterator(AsyncIterator["PublishedFile"]):
    def __init__(
        self,
        state: ConnectionState,
        user: BaseUser,
        game: Game | None,
        type: PublishedFileType = PublishedFileType.Community,
        revision: PublishedFileRevision = PublishedFileRevision.Default,
        language: Language | None = None,
        limit: int | None = None,
        before: datetime | None = None,
        after: datetime | None = None,
    ):
        super().__init__(state, limit, before, after)
        self.user = user
        self.app_id = getattr(game, "id", 0)
        self.revision = revision
        self.type = type
        self.language = language

    async def fill(self):
        from .published_file import PublishedFile

        initial = await self._state.fetch_user_published_files(
            self.user.id64, self.app_id, 1, self.type, self.revision, self.language
        )

        for file in initial.publishedfiledetails:
            file = PublishedFile(self._state, file, self.user)
            if not self.after < file.created_at < self.before:
                return
            yield file

        for page in range(2, math.ceil((initial.total + 30) / 30)):
            msg = await self._state.fetch_user_published_files(
                self.user.id64, self.app_id, page, self.type, self.revision, self.language
            )

            for file in msg.publishedfiledetails:
                file = PublishedFile(self._state, file, self.user)
                if not self.after < file.created_at < self.before:
                    return
                yield file


class GamePublishedFilesIterator(AsyncIterator["PublishedFile"]):
    def __init__(
        self,
        state: ConnectionState,
        game: Game,
        type: PublishedFileQueryFileType = PublishedFileQueryFileType.Items,
        revision: PublishedFileRevision = PublishedFileRevision.Default,
        language: Language | None = None,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ):
        super().__init__(state, limit, before, after)
        self.game = game
        self.type = type
        self.revision = revision
        self.language = language

    async def fill(self):
        from .published_file import PublishedFile

        remaining = None
        cursor = "*"

        while remaining is None or remaining > 0:
            protos = await self._state.fetch_game_published_files(
                self.game.id, self.after, self.before, self.type, self.revision, self.language, self.limit, cursor
            )
            if remaining is None:
                remaining = protos.total
            remaining -= len(protos.publishedfiledetails)
            cursor = protos.next_cursor

            files: list[PublishedFile] = []
            for file in protos.publishedfiledetails:
                author: Authors = file.creator  # type: ignore
                file = PublishedFile(self._state, file, author)
                if not self.after < file.created_at < self.before:
                    remaining = 0
                    break
                files.append(file)

            await self._fill_queue_users(files)
            for file in files:
                yield file



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import abc
import asyncio
import logging
import re
import traceback
from collections.abc import AsyncGenerator, Callable
from contextlib import asynccontextmanager
from datetime import timedelta
from io import BytesIO
from typing import TYPE_CHECKING, Any, TypeVar

from typing_extensions import Literal, ParamSpec, Self, TypedDict
from yarl import URL as _URL

from . import utils
from ._const import TASK_HAS_NAME, URL
from .enums import IntEnum
from .image import Image
from .protobufs import EMsg

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath

    from .gateway import Msgs
    from .state import ConnectionState


__all__ = (
    "PriceOverview",
    "Ban",
)

F = TypeVar("F", bound="Callable[..., Any]")
P = ParamSpec("P")


def api_route(path: str, version: int = 1) -> _URL:
    return URL.API / f"{path}/v{version}"


class _ReturnTrue:
    __slots__ = ()

    def __call__(self, *_: Any, **__: Any) -> Literal[True]:
        return True

    def __repr__(self) -> str:
        return "<return_true>"


return_true = _ReturnTrue()


class Registerable:
    __slots__ = ("loop", "parsers_name")
    loop: asyncio.AbstractEventLoop

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        self = super().__new__(cls)
        self.loop = asyncio.get_event_loop()
        cls.parsers_name = tuple(cls.__annotations__)[0]
        bases = tuple(reversed(cls.__mro__[:-2]))  # skip Registerable and object
        for idx, base in enumerate(bases):
            parsers_name = tuple(base.__annotations__)[0]
            for name, attr in base.__dict__.items():
                if not hasattr(attr, "msg"):
                    continue
                try:
                    parsers = getattr(self, parsers_name)
                except AttributeError:
                    parsers = {}
                    setattr(self, parsers_name, parsers)
                msg_parser = getattr(self, name)
                if idx != 0 and isinstance(attr.msg, EMsg):
                    parsers = getattr(self, tuple(bases[0].__annotations__)[0])
                parsers[attr.msg] = msg_parser

        return self

    @staticmethod
    def _run_parser_callback(task: asyncio.Task[object]) -> None:
        try:
            exception = task.exception()
        except asyncio.CancelledError:
            return
        if exception:
            traceback.print_exception(exception.__class__, exception, exception.__traceback__)

    def run_parser(self, emsg: IntEnum, msg: Msgs) -> None:
        try:
            event_parser = getattr(self, self.parsers_name)[emsg]
        except KeyError:
            log = logging.getLogger(self.__class__.__module__)
            try:
                log.debug("Ignoring event %r", msg)
            except Exception:
                log.debug("Ignoring event %s", msg.msg)
        else:
            coro = utils.maybe_coroutine(event_parser, msg)

            (
                self.loop.create_task(coro, name=f"task_{event_parser.__name__}")
                if TASK_HAS_NAME
                else self.loop.create_task(coro)
            ).add_done_callback(self._run_parser_callback)


EventParser = None


def register(msg: IntEnum) -> Callable[[F], F]:  # this afaict is not type able currently without HKT
    def wrapper(callback: F) -> F:
        callback.msg = msg
        return callback

    return wrapper


PRICE_RE = re.compile(r"(^\D*(?P<price>[\d,.]*)\D*$)")


class PriceOverviewDict(TypedDict):
    success: bool
    lowest_price: str
    median_price: str
    volume: str


class PriceOverview:
    """Represents the data received from https://steamcommunity.com/market/priceoverview.

    Attributes
    -------------
    currency
        The currency identifier for the item e.g. "$" or "£".
    volume
        The amount of items that sold in the last 24 hours.
    lowest_price
        The lowest price observed by the market.
    median_price
        The median price observed by the market.
    """

    __slots__ = ("currency", "volume", "lowest_price", "median_price")

    lowest_price: float | str
    median_price: float | str

    def __init__(self, data: PriceOverviewDict):
        lowest_price = PRICE_RE.search(data["lowest_price"])["price"]
        median_price = PRICE_RE.search(data["median_price"])["price"]

        try:
            self.lowest_price = float(lowest_price.replace(",", "."))
            self.median_price = float(median_price.replace(",", "."))
        except ValueError:
            self.lowest_price = lowest_price
            self.median_price = median_price

        self.volume: int = int(data["volume"].replace(",", ""))
        self.currency: str = data["lowest_price"].replace(lowest_price, "").strip()

    def __repr__(self) -> str:
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in self.__slots__]
        return f"<PriceOverview {' '.join(resolved)}>"


class Ban:
    """Represents a Steam ban.

    Attributes
    ----------
    since_last_ban
        How many days since the user was last banned
    number_of_game_bans
        The number of game bans the user has.
    """

    __slots__ = (
        "since_last_ban",
        "number_of_game_bans",
        "_vac_banned",
        "_community_banned",
        "_market_banned",
    )

    def __init__(self, data: dict[str, Any]):
        self._vac_banned: bool = data["VACBanned"]
        self._community_banned: bool = data["CommunityBanned"]
        self._market_banned: bool = data["EconomyBan"]
        self.since_last_ban = timedelta(days=data["DaysSinceLastBan"])
        self.number_of_game_bans: int = data["NumberOfGameBans"]

    def __repr__(self) -> str:
        attrs = [
            ("number_of_game_bans", self.number_of_game_bans),
            ("is_vac_banned()", self.is_vac_banned()),
            ("is_community_banned()", self.is_community_banned()),
            ("is_market_banned()", self.is_market_banned()),
        ]
        resolved = [f"{method}={value!r}" for method, value in attrs]
        return f"<Ban {' '.join(resolved)}>"

    def is_banned(self) -> bool:
        """Species if the user is banned from any part of Steam."""
        return any((self.is_vac_banned(), self.is_community_banned(), self.is_market_banned()))

    def is_vac_banned(self) -> bool:
        """Whether or not the user is VAC banned."""
        return self._vac_banned

    def is_community_banned(self) -> bool:
        """Whether or not the user is community banned."""
        return self._community_banned

    def is_market_banned(self) -> bool:
        """Whether or not the user is market banned."""
        return self._market_banned


class _IOMixin(metaclass=abc.ABCMeta):
    __slots__ = ()

    @asynccontextmanager
    async def open(self, **kwargs: Any) -> AsyncGenerator[BytesIO, None]:
        """Open this file as and returns its contents as an in memory buffer."""
        try:
            url: str = self.url  # type: ignore
            state: ConnectionState = self._state  # type: ignore
        except AttributeError:
            raise NotImplementedError() from None

        async with state.http._session.get(url) as r:
            yield BytesIO(await r.read())

    async def read(self, **kwargs: Any) -> bytes:
        """Read the whole contents of this file."""
        async with self.open(**kwargs) as io:
            return io.getvalue()

    async def save(self, filename: StrOrBytesPath, **kwargs: Any) -> int:
        """Save the file to a path.

        Parameters
        ----------
        filename
            The filename of the file to be created and have this saved to.

        Returns
        -------
        The number of bytes written.
        """
        async with self.open(**kwargs) as file:
            with file, open(filename, "wb") as actual_fp:
                return actual_fp.write(file.getvalue())

    async def image(self, *, spoiler: bool = False, **kwargs: Any) -> Image:
        """Return this file as an image for uploading."""
        async with self.open(**kwargs) as file:
            return Image(file, spoiler=spoiler)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

from multidict import MultiDict
from typing_extensions import Never, TypeAlias, TypedDict

# VDF related types
try:
    import orvdf
except ModuleNotFoundError:
    VDFDict: TypeAlias = "MultiDict[str | VDFDict]"
    BinaryVDFDict: TypeAlias = "MultiDict[str | int | float | BinaryVDFDict]"
else:
    VDFDict: TypeAlias = "MultiDict[str | bool | None | VDFDict]"
    BinaryVDFDict: TypeAlias = "MultiDict[str | int | float | bool | None | BinaryVDFDict]"


T = TypeVar("T")
VDFInt: TypeAlias = str


class VDFList(MultiDict[T]):
    # lists don't strictly exist in the VDF "spec" however Valve uses {"0": T, "1": T, ..., "n": T}
    # to represent them, so prevent people doing daft things like use .keys() and iterating over it
    # for now using __getitem__ is fine, but I might remove this at some point

    def keys(self) -> Never:
        ...

    def __iter__(self) -> Never:
        ...


if TYPE_CHECKING:

    class TypedVDFDict(MultiDict[Any], TypedDict):  # type: ignore  # this obviously doesn't work at runtime
        pass

else:

    class TypedVDFDict(MultiDict[Any]):
        def __init_subclass__(cls, **kwargs: Any) -> None:
            pass



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator, Coroutine
from contextlib import asynccontextmanager
from datetime import datetime
from typing import TYPE_CHECKING, Any

from .abc import Channel
from .chat import Chat, GroupChannelProtos
from .iterators import DMChannelHistoryIterator
from .message import ClanMessage, GroupMessage, UserMessage

if TYPE_CHECKING:
    from .clan import Clan
    from .group import Group
    from .image import Image
    from .state import ConnectionState
    from .user import User

__all__ = (
    "DMChannel",
    "GroupChannel",
    "ClanChannel",
)


class DMChannel(Channel[UserMessage]):  # TODO cache these to add last_message.
    """Represents the channel a DM is sent in.

    Attributes
    ----------
    participant
        The recipient of any messages sent.
    """

    __slots__ = ("participant",)
    clan: None
    group: None

    def __init__(self, state: ConnectionState, participant: User):
        super().__init__(state)
        self.participant = participant
        self.clan = None
        self.group = None

    def __repr__(self) -> str:
        return f"<DMChannel participant={self.participant!r}>"

    def __eq__(self, other: object) -> bool:
        return self.participant == other.participant if isinstance(other, DMChannel) else NotImplemented

    def _message_func(self, content: str) -> Coroutine[Any, Any, UserMessage]:
        return self.participant._message_func(content)

    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        return self.participant._image_func(image)

    @asynccontextmanager
    async def typing(self) -> AsyncGenerator[None, None]:
        """Send a typing indicator continuously to the channel while in the context manager.

        Note
        ----
        This only works in DMs.

        Usage:

        .. code-block:: python3

            async with channel.typing():
                ...  # do your expensive operations
        """

        async def inner() -> None:
            while True:
                await asyncio.sleep(10)
                await self.trigger_typing()

        await self.trigger_typing()
        task = self._state.loop.create_task(inner())
        yield
        task.cancel()

    async def trigger_typing(self) -> None:
        """Send a typing indicator to the channel once.

        Note
        ----
        This only works in DMs.
        """
        await self._state.send_user_typing(self.participant.id64)

    def history(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> DMChannelHistoryIterator:
        return DMChannelHistoryIterator(state=self._state, channel=self, limit=limit, before=before, after=after)


class GroupChannel(Chat[GroupMessage]):
    """Represents a group channel.

    Attributes
    ----------
    id
        The ID of the channel.
    name
        The name of the channel, this could be the same as the :attr:`~steam.Group.name` if it's the main channel.
    group
        The group to which messages are sent.
    joined_at
        The time the client joined the chat.
    position
        The position of the channel in the channel list.
    last_message
        The last message sent in the channel.
    """

    clan: None

    def __init__(self, state: ConnectionState, group: Group, proto: GroupChannelProtos):
        super().__init__(state, group, proto)
        self.group: Group = group

    def _message_func(self, content: str) -> Coroutine[Any, Any, GroupMessage]:
        return self._state.send_chat_message(*self._location, content)  # type: ignore

    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        return self._state.http.send_chat_image(*self._location, image)


class ClanChannel(Chat[ClanMessage]):
    """Represents a group channel.

    Attributes
    ----------
    id
        The ID of the channel.
    name
        The name of the channel, this could be the same as the :attr:`~steam.Clan.name` if it's the main channel.
    clan
        The clan to which messages are sent.
    joined_at
        The time the client joined the chat.
    position
        The position of the channel in the channel list.
    last_message
        The last message sent in the channel.
    """

    group: None

    def __init__(self, state: ConnectionState, clan: Clan, proto: GroupChannelProtos):
        super().__init__(state, clan, proto)
        self.clan: Clan = clan

    def _message_func(self, content: str) -> Coroutine[Any, Any, ClanMessage]:
        return self._state.send_chat_message(*self._location, content)  # type: ignore

    def _image_func(self, image: Image) -> Coroutine[Any, Any, None]:
        return self._state.http.send_chat_image(*self._location, image)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import inspect
import sys
import traceback
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, overload

from typing_extensions import Final

from ... import ClientException
from .commands import Command, Group

if TYPE_CHECKING:
    from steam.ext import commands

    from ...client import E, EventType
    from .bot import Bot

__all__ = ("Cog",)


class Cog:
    """A class from which Cogs can be created. These are used to separate commands and listeners into separate files.

    Attributes
    ----------
    qualified_name
        The name of the cog. Can be set in a subclass e.g.

        .. code-block:: python3

            class MyCog(commands.Cog, name="SpecialCogName"):
                ...

        Defaults to ``Cog.__name__``.

    command_attrs
        Attributes to pass to every command registered in the cog.
        Can be set in subclass e.g.

        .. code-block:: python3

            class MyBrokenCog(commands.Cog, command_attrs=dict(enabled=False)):
                # all the commands by default would be disabled

                @commands.command()
                async def broken(self, ctx):  # disabled
                    ...

                @commands.command(enabled=True)
                async def working(self, ctx):  # enabled
                    ...

    description
        The cleaned up doc-string for the cog.
    """

    __commands__: Final[dict[str, Command]] = ...
    __listeners__: Final[dict[str, list[EventType]]] = ...
    command_attrs: Final[dict[str, Any]] = ...
    qualified_name: Final[str] = ...
    description: Final[str | None] = ...

    def __init_subclass__(cls, name: str | None = None, command_attrs: dict[str, Any] | None = None) -> None:
        cls.qualified_name = name or cls.__name__  # type: ignore
        cls.command_attrs = command_attrs or {}  # type: ignore

        if cls.__doc__ is not None:
            cls.description = inspect.cleandoc(cls.__doc__)  # type: ignore
        else:
            cls.description = None  # type: ignore

        cls.__listeners__ = {}  # type: ignore
        cls.__commands__ = {}  # type: ignore
        for name, attr in inspect.getmembers(cls):
            if name.startswith(("bot_", "cog_")) and getattr(Cog, name, None) is None:
                raise ClientException(
                    f'Methods prefixed with "bot_" or "cog_" are reserved for future use, {attr.__qualname__} is '
                    f"therefore not allowed"
                )
            if isinstance(attr, Command):
                if isinstance(attr, Group):
                    for name, value in cls.command_attrs.items():
                        for child in attr.children:
                            if name not in child.__original_kwargs__:
                                setattr(child, name, value)
                else:
                    cls.__commands__[name] = attr

                for name, value in cls.command_attrs.items():
                    if name not in attr.__original_kwargs__:
                        setattr(attr, name, value)

            elif hasattr(attr, "__event_name__"):
                try:
                    cls.__listeners__[attr.__event_name__].append(attr)
                except KeyError:
                    cls.__listeners__[attr.__event_name__] = [attr]

    def __repr__(self) -> str:
        return f"<Cog {f'{self.__class__.__module__}.{self.__class__.__name__}'!r}>"

    # TODO for 3.9 make these class properties
    @property
    def commands(self) -> set[Command]:
        """A set of the cog's commands."""
        return set(self.__commands__.values())

    @property
    def listeners(self) -> list[tuple[str, EventType]]:
        """A list of tuples of the events registered with the format (name, listener)"""
        return [(listener.__name__, listener) for listeners in self.__listeners__.values() for listener in listeners]

    @classmethod
    @overload
    def listener(cls, coro: E) -> E:
        ...

    @classmethod
    @overload
    def listener(cls, name: str | None = None) -> Callable[[E], E]:
        ...

    @classmethod
    def listener(cls, name: E | str | None = None) -> Callable[[E], E]:
        """|maybecallabledeco|
        Register a :term:`coroutine function` as a listener. Similar to :meth:`~steam.ext.commands.Bot.listen`.

        Parameters
        ----------
        name: :class:`str`
            The name of the event to listen for. Defaults to ``func.__name__``.
        """

        def decorator(coro: E) -> E:
            if not inspect.iscoroutinefunction(coro):
                raise TypeError(f"Listeners must be coroutine functions, {coro.__name__} is {type(coro).__name__}")
            coro.__event_name__ = name if not callable(name) else coro.__name__
            return coro

        return decorator(name) if callable(name) else decorator

    async def cog_command_error(self, ctx: "commands.Context", error: Exception) -> None:
        """A special method that is called when an error is dispatched inside this cog. This is similar to
        :func:`~commands.Bot.on_command_error` except it only applies to the commands inside this cog.

        Parameters
        -----------
        ctx
            The invocation context where the error happened.
        error
            The error that happened.
        """
        print(f"Ignoring exception in command {ctx.command}:", file=sys.stderr)
        traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)

    async def cog_check(self, ctx: "commands.Context") -> bool:
        """|maybecoro|
        A special method that registers as a :func:`commands.check` for every command and subcommand in this cog.
        This should return a boolean result.

        Parameters
        -----------
        ctx
            The invocation context.
        """

    async def bot_check(self, ctx: "commands.Context") -> bool:
        """|maybecoro|
        A special method that registers as a :meth:`commands.Bot.check` globally. This should return a boolean result.

        Parameters
        -----------
        ctx
            The invocation context.
        """

    def cog_unload(self) -> None:
        """A special method that is called when the cog gets removed.

        This is called before :func:`teardown`.
        """

    def _inject(self, bot: Bot) -> None:
        cls = self.__class__
        if cls.bot_check is not Cog.bot_check:
            bot.add_check(self.bot_check)

        for idx, command in enumerate(self.__commands__.values()):
            command.cog = self

            if cls.cog_check is not Cog.cog_check:
                command.checks.append(self.cog_check)

            if isinstance(command, Group):
                for child in command.children:
                    del child.clean_params
                    child.cog = self
                    child.clean_params

            for decorator in Command.DECORATORS:
                command_deco = getattr(command, decorator, None)
                if command_deco is not None:  # update any unbound decorators
                    setattr(command, command_deco.__name__, getattr(self, command_deco.__name__))

            try:
                bot.add_command(command)
            except Exception:
                # undo our additions
                for to_undo in tuple(self.__commands__)[:idx]:
                    bot.remove_command(to_undo)
                raise

        for name, listeners in self.__listeners__.items():
            for listener in listeners:
                listener = getattr(self, listener.__name__)  # get the bound method version
                bot.add_listener(listener, name)

    def _eject(self, bot: Bot) -> None:
        for command in self.__commands__.values():
            if isinstance(command, Group):
                command.recursively_remove_all_commands()
            else:
                bot.remove_command(command.name)

        for name, listeners in self.__listeners__.items():
            for listener in listeners:
                listener = getattr(self, listener.__name__)  # get the bound method version
                bot.remove_listener(listener, name)

        self.cog_unload()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_gameservers.proto
# plugin: python-betterproto
# Last updated 18/05/2022

from dataclasses import dataclass
from typing import Dict, List

import betterproto

from .base import CMsgIpAddress


class EQueryType(betterproto.Enum):
    Invalid = 0
    Ping = 1
    Players = 2
    Rules = 3


@dataclass(eq=False, repr=False)
class GetServerListRequest(betterproto.Message):
    filter: str = betterproto.string_field(1)
    limit: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class GetServerListResponse(betterproto.Message):
    servers: List["GetServerListResponseServer"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetServerListResponseServer(betterproto.Message):
    addr: str = betterproto.string_field(1)
    gameport: int = betterproto.uint32_field(2)
    specport: int = betterproto.uint32_field(3)
    steamid: int = betterproto.fixed64_field(4)
    name: str = betterproto.string_field(5)
    appid: int = betterproto.uint32_field(6)
    gamedir: str = betterproto.string_field(7)
    version: str = betterproto.string_field(8)
    product: str = betterproto.string_field(9)
    region: int = betterproto.int32_field(10)
    players: int = betterproto.int32_field(11)
    max_players: int = betterproto.int32_field(12)
    bots: int = betterproto.int32_field(13)
    map: str = betterproto.string_field(14)
    secure: bool = betterproto.bool_field(15)
    dedicated: bool = betterproto.bool_field(16)
    os: str = betterproto.string_field(17)
    gametype: str = betterproto.string_field(18)


@dataclass(eq=False, repr=False)
class GetServerSteamIDsByIpRequest(betterproto.Message):
    server_ips: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class IPsWithSteamIDsResponse(betterproto.Message):
    servers: List["IPsWithSteamIDsResponseServer"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class IPsWithSteamIDsResponseServer(betterproto.Message):
    addr: str = betterproto.string_field(1)
    steamid: int = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class GetServerIPsBySteamIdRequest(betterproto.Message):
    server_steamids: List[int] = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class QueryRequest(betterproto.Message):
    fake_ip: int = betterproto.uint32_field(1)
    fake_port: int = betterproto.uint32_field(2)
    app_id: int = betterproto.uint32_field(3)
    query_type: EQueryType = betterproto.enum_field(4)


@dataclass(eq=False, repr=False)
class PingQueryData(betterproto.Message):
    server_ip: CMsgIpAddress = betterproto.message_field(1)
    query_port: int = betterproto.uint32_field(2)
    game_port: int = betterproto.uint32_field(3)
    spectator_port: int = betterproto.uint32_field(4)
    spectator_server_name: str = betterproto.string_field(5)
    server_name: str = betterproto.string_field(6)
    steamid: int = betterproto.fixed64_field(7)
    app_id: int = betterproto.uint32_field(8)
    gamedir: str = betterproto.string_field(9)
    map: str = betterproto.string_field(10)
    game_description: str = betterproto.string_field(11)
    gametype: str = betterproto.string_field(12)
    num_players: int = betterproto.uint32_field(13)
    max_players: int = betterproto.uint32_field(14)
    num_bots: int = betterproto.uint32_field(15)
    password: bool = betterproto.bool_field(16)
    secure: bool = betterproto.bool_field(17)
    dedicated: bool = betterproto.bool_field(18)
    version: str = betterproto.string_field(19)
    sdr_popid: int = betterproto.fixed32_field(20)
    sdr_location_string: str = betterproto.string_field(21)


@dataclass(eq=False, repr=False)
class PlayersQueryDataPlayer(betterproto.Message):
    name: str = betterproto.string_field(1)
    score: int = betterproto.uint32_field(2)
    time_played: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class PlayersQueryData(betterproto.Message):
    players: List[PlayersQueryDataPlayer] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class RulesQueryData(betterproto.Message):
    rules: Dict[str, str] = betterproto.map_field(1, betterproto.TYPE_STRING, betterproto.TYPE_STRING)


@dataclass(eq=False, repr=False)
class QueryResponse(betterproto.Message):
    ping_data: PingQueryData = betterproto.message_field(1)
    players_data: PlayersQueryData = betterproto.message_field(2)
    rules_data: RulesQueryData = betterproto.message_field(3)
